from .._discrete_distribution import LD
from ...util import ParameterError, ParameterWarning
import qmctoolscl
from os.path import dirname, abspath, isfile
from numpy import *
from numpy.lib.npyio import DataSource
import warnings


class DigitalNetB2(LD):
    """
    Quasi-Random digital nets in base 2.
    
    >>> dnb2 = DigitalNetB2(2,seed=7)
    >>> dnb2.gen_samples(4)
    array([[0.0715562 , 0.07784108],
           [0.81420169, 0.74485558],
           [0.31409299, 0.93233913],
           [0.57163057, 0.26535753]])
    >>> dnb2.gen_samples(1)
    array([[0.0715562 , 0.07784108]])
    >>> dnb2
    DigitalNetB2 (DiscreteDistribution Object)
        d               2^(1)
        dvec            [0 1]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()
    >>> DigitalNetB2(dimension=2,randomize=False,graycode=True).gen_samples(n_min=2,n_max=4,warn=False)
    array([[0.75, 0.25],
           [0.25, 0.75]])
    >>> DigitalNetB2(dimension=2,randomize=False,graycode=False).gen_samples(n_min=2,n_max=4,warn=False)
    array([[0.25, 0.75],
           [0.75, 0.25]])
    >>> dnb2_alpha2 = DigitalNetB2(5,randomize=False,generating_matrices='sobol_mat_alpha2.10600.64.32.lsb.npy')
    >>> dnb2_alpha2.gen_samples(8,warn=False)
    array([[0.      , 0.      , 0.      , 0.      , 0.      ],
           [0.75    , 0.75    , 0.75    , 0.75    , 0.75    ],
           [0.4375  , 0.9375  , 0.1875  , 0.6875  , 0.1875  ],
           [0.6875  , 0.1875  , 0.9375  , 0.4375  , 0.9375  ],
           [0.296875, 0.171875, 0.109375, 0.796875, 0.859375],
           [0.546875, 0.921875, 0.859375, 0.046875, 0.109375],
           [0.234375, 0.859375, 0.171875, 0.484375, 0.921875],
           [0.984375, 0.109375, 0.921875, 0.734375, 0.171875]])
    >>> DigitalNetB2(dimension=3,randomize=False,generating_matrices="LDData/main/dnet/mps.nx_s5_alpha2_m32.txt").gen_samples(8,warn=False)
    array([[0.        , 0.        , 0.        ],
           [0.75841841, 0.45284834, 0.48844557],
           [0.57679828, 0.13226272, 0.10061957],
           [0.31858402, 0.32113875, 0.39369111],
           [0.90278927, 0.45867532, 0.01803333],
           [0.14542431, 0.02548793, 0.4749614 ],
           [0.45587539, 0.33081476, 0.11474426],
           [0.71318879, 0.15377192, 0.37629925]])
    >>> DigitalNetB2(dimension=3,randomize='LMS_DS',seed=5).gen_samples(8)
    array([[0.87245179, 0.63803439, 0.06539576],
           [0.37498111, 0.49503961, 0.57038663],
           [0.56914076, 0.10224734, 0.98079258],
           [0.06672611, 0.78027648, 0.47528302],
           [0.87882949, 0.26075657, 0.3249126 ],
           [0.37642258, 0.62176726, 0.81867062],
           [0.67994388, 0.98330928, 0.72078745],
           [0.18248071, 0.14976471, 0.22748731]])
    >>> DigitalNetB2(dimension=3,randomize='DS',seed=5).gen_samples(8)
    array([[0.68383949, 0.04047995, 0.42903182],
           [0.18383949, 0.54047995, 0.92903182],
           [0.93383949, 0.79047995, 0.67903182],
           [0.43383949, 0.29047995, 0.17903182],
           [0.55883949, 0.66547995, 0.05403182],
           [0.05883949, 0.16547995, 0.55403182],
           [0.80883949, 0.41547995, 0.80403182],
           [0.30883949, 0.91547995, 0.30403182]])
    >>> DigitalNetB2(dimension=3,randomize='OWEN',seed=5).gen_samples(8)
    array([[0.33595486, 0.05834975, 0.30066401],
           [0.89110875, 0.84905188, 0.81833285],
           [0.06846074, 0.59997956, 0.67064205],
           [0.6693703 , 0.25824002, 0.10469644],
           [0.44586618, 0.99161977, 0.1873488 ],
           [0.84245267, 0.16445553, 0.56544372],
           [0.18546359, 0.44859876, 0.97389524],
           [0.61215442, 0.64341386, 0.44529863]])
    >>> DigitalNetB2(dimension=3,randomize='LMS',seed=5).gen_samples(8,warn=False)
    array([[0.        , 0.        , 0.        ],
           [0.50255985, 0.86689489, 0.51063711],
           [0.30826124, 0.72453078, 0.92054737],
           [0.80587711, 0.39291452, 0.41042919],
           [0.24859572, 0.8812071 , 0.26392735],
           [0.74615831, 0.23623819, 0.75454159],
           [0.44253238, 0.34735206, 0.65710831],
           [0.9450387 , 0.51954284, 0.16603643]])
    >>> DigitalNetB2(dimension=3,randomize='LMS_DS',seed=5,alpha=2).gen_samples(8)
    array([[0.24801998, 0.43503469, 0.32346891],
           [0.91475727, 0.68371042, 0.60623725],
           [0.34781088, 0.60489022, 0.39960482],
           [0.54824202, 0.35357466, 0.68627971],
           [0.39739254, 0.25987535, 0.29547912],
           [0.6910429 , 0.51121493, 0.50915542],
           [0.01488653, 0.71489405, 0.49668689],
           [0.83832067, 0.4662247 , 0.71426927]])
    >>> DigitalNetB2(dimension=3,randomize='LMS',seed=5,alpha=2).gen_samples(8,warn=False)
    array([[0.        , 0.        , 0.        ],
           [0.83330849, 0.75137017, 0.78900948],
           [0.40020937, 0.95905275, 0.20511411],
           [0.699824  , 0.21041396, 0.99021704],
           [0.35457947, 0.17909801, 0.09927936],
           [0.56088333, 0.92774329, 0.8148018 ],
           [0.23708574, 0.84517128, 0.17932156],
           [0.91360582, 0.09382525, 0.89093793]])
    >>> DigitalNetB2(dimension=3,randomize='False',seed=5,alpha=2).gen_samples(8,warn=False)
    array([[0.      , 0.      , 0.      ],
           [0.75    , 0.75    , 0.75    ],
           [0.4375  , 0.9375  , 0.1875  ],
           [0.6875  , 0.1875  , 0.9375  ],
           [0.296875, 0.171875, 0.109375],
           [0.546875, 0.921875, 0.859375],
           [0.234375, 0.859375, 0.171875],
           [0.984375, 0.109375, 0.921875]])
    >>> DigitalNetB2(dimension=3,randomize='NUS',seed=5,alpha=2).gen_samples(8)
    array([[0.13470966, 0.19119158, 0.96130884],
           [0.97413754, 0.90343374, 0.02884907],
           [0.26275262, 0.86277465, 0.75808903],
           [0.59591889, 0.0264989 , 0.21845493],
           [0.49745711, 0.1158892 , 0.92224289],
           [0.65012418, 0.75808168, 0.09617135],
           [0.11594365, 0.98245561, 0.850095  ],
           [0.7761354 , 0.16906349, 0.15959343]])
    >>> DigitalNetB2(dimension=3,randomize='LMS_DS',seed=7,replications=2).gen_samples(4)
    array([[[0.31330935, 0.09580848, 0.24636182],
            [0.61112103, 0.66930206, 0.70007472],
            [0.10167532, 0.91581583, 0.97325936],
            [0.83691425, 0.35015003, 0.45724578]],
    <BLANKLINE>
           [[0.13239161, 0.18660628, 0.73760357],
            [0.76777532, 0.52700602, 0.17183852],
            [0.39365722, 0.91359273, 0.43231666],
            [0.50656501, 0.25313607, 0.97464392]]])
    >>> DigitalNetB2(dimension=3,randomize='LMS',seed=7,replications=2).gen_samples(4,warn=False)
    array([[[0.        , 0.        , 0.        ],
            [0.79796457, 0.70241587, 0.5475088 ],
            [0.28983202, 0.94903657, 0.7742629 ],
            [0.52522336, 0.254479  , 0.28954498]],
    <BLANKLINE>
           [[0.        , 0.        , 0.        ],
            [0.8961262 , 0.66083813, 0.59054646],
            [0.27006262, 0.77399226, 0.82226327],
            [0.62613416, 0.43372985, 0.27077997]]])
    >>> DigitalNetB2(dimension=3,randomize='DS',seed=7,replications=2).gen_samples(4)
    array([[[0.04386058, 0.58727432, 0.3691824 ],
            [0.54386058, 0.08727432, 0.8691824 ],
            [0.29386058, 0.33727432, 0.6191824 ],
            [0.79386058, 0.83727432, 0.1191824 ]],
    <BLANKLINE>
           [[0.65212985, 0.69669968, 0.10605352],
            [0.15212985, 0.19669968, 0.60605352],
            [0.90212985, 0.44669968, 0.85605352],
            [0.40212985, 0.94669968, 0.35605352]]])
    >>> DigitalNetB2(dimension=3,randomize='OWEN',seed=7,replications=2).gen_samples(4)
    array([[[0.35967208, 0.89954733, 0.20817919],
            [0.98911419, 0.49964829, 0.62480864],
            [0.02600317, 0.16945373, 0.97946126],
            [0.56008744, 0.54552841, 0.30711945]],
    <BLANKLINE>
           [[0.00782812, 0.71181447, 0.44244982],
            [0.53880628, 0.41910168, 0.65404289],
            [0.29051333, 0.16752862, 0.80878263],
            [0.77449486, 0.83546482, 0.22950922]]])
    
    References:

        [1] Marius Hofert and Christiane Lemieux (2019). 
        qrng: (Randomized) Quasi-Random Number Generators. 
        R package version 0.0-7.
        https://CRAN.R-project.org/package=qrng.

        [2] Faure, Henri, and Christiane Lemieux. 
        “Implementation of Irreducible Sobol' Sequences in Prime Power Bases.” 
        Mathematics and Computers in Simulation 161 (2019): 13-22. Crossref. Web.

        [3] F.Y. Kuo & D. Nuyens.
        Application of quasi-Monte Carlo methods to elliptic PDEs with random diffusion coefficients 
        - a survey of analysis and implementation, Foundations of Computational Mathematics, 
        16(6):1631-1696, 2016.
        springer link: https://link.springer.com/article/10.1007/s10208-016-9329-5
        arxiv link: https://arxiv.org/abs/1606.06613
        
        [4] D. Nuyens, `The Magic Point Shop of QMC point generators and generating
        vectors.` MATLAB and Python software, 2018. Available from
        https://people.cs.kuleuven.be/~dirk.nuyens/
        https://bitbucket.org/dnuyens/qmc-generators/src/cb0f2fb10fa9c9f2665e41419097781b611daa1e/cpp/digitalseq_b2g.hpp

        [5] Paszke, A., Gross, S., Massa, F., Lerer, A., Bradbury, J., Chanan, G., … Chintala, S. 
        (2019). PyTorch: An Imperative Style, High-Performance Deep Learning Library. 
        In H. Wallach, H. Larochelle, A. Beygelzimer, F. d extquotesingle Alch&#39;e-Buc, E. Fox, & R. Garnett (Eds.), 
        Advances in Neural Information Processing Systems 32 (pp. 8024-8035). Curran Associates, Inc. 
        Retrieved from http://papers.neurips.cc/paper/9015-pytorch-an-imperative-style-high-performance-deep-learning-library.pdf

        [6] I.M. Sobol', V.I. Turchaninov, Yu.L. Levitan, B.V. Shukhman: 
        "Quasi-Random Sequence Generators" Keldysh Institute of Applied Mathematics, 
        Russian Academy of Sciences, Moscow (1992).

        [7] Sobol, Ilya & Asotsky, Danil & Kreinin, Alexander & Kucherenko, Sergei. (2011). 
        Construction and Comparison of High-Dimensional Sobol' Generators. Wilmott. 
        2011. 10.1002/wilm.10056. 

        [8] Paul Bratley and Bennett L. Fox. 1988. 
        Algorithm 659: Implementing Sobol's quasirandom sequence generator. 
        ACM Trans. Math. Softw. 14, 1 (March 1988), 88-100. 
        DOI:https://doi.org/10.1145/42288.214372
    """

    def __init__(self, dimension=1, randomize='LMS_DS', graycode=False, seed=None, 
        generating_matrices='sobol_mat.21201.32.32.msb.npy', d_max=None, t_max=None, m_max=None, msb=None, t_lms=53, alpha=1,
        _verbose=False, replications=1, qmctoolscl_kwargs={"backend":"c"}):
        r"""
        Args:
            dimension (int or :class:`numpy.ndarray`): dimension of the generator. 
                
                - If an int is passed in, use sequence dimensions [0,...,dimension-1].
                - If an ndarry is passed in, use these dimension indices in the sequence. 

            randomize (bool): apply randomization? True defaults to LMS_DS. Can also explicitly pass in
                
                - "LMS_DS": linear matrix scramble with digital shift
                - "LMS": linear matrix scramble only
                - "DS": digital shift only
                - "OWEN" or "NUS": nested uniform scrambling (Owen scrambling)

            graycode (bool): indicator to use graycode ordering (True) or natural ordering (False)
            seed (list): int seed of list of seeds, one for each dimension.
            generating_matrices (:class:`numpy.ndarray` or str): Specify generating matrices. There are a number of optional input types. 
                
                - An ndarray of integers with shape (`d_max`, `m_max)` where each int has `t_max` bits.
                - A string with either a relative path from `LDData repository <https://github.com/QMCSoftware/LDData>`__ (e.g., "LDData/main/dnet/mps.nx_s5_alpha3_m32.txt") or a NumPy file with format "name.d_max.t_max.m_max.{msb,lsb}.npy" (e.g., "gen_mat.21201.32.32.msb.npy")

            d_max (int): max dimension
            t_max (int): number of bits in each int of each generating matrix, aka: number of rows in a generating matrix with ints expanded into columns
            m_max (int): :math:`2^{\\texttt{m\_max}}` is the number of samples supported, aka: number of columns in a generating matrix with ints expanded into columns
            msb (bool): bit storage as ints, e.g., if :math:`{\\texttt{t\_max}} = 3`, then 6 is [1 1 0] in MSB (True) and [0 1 1] in LSB (False)
            alpha (int): interlacing factor for higher order nets. 
                WARNING: When alpha>1, interlacing is performed regardless of the generating matrices 
                i.e. for alpha>1 do NOT pass in generating matrices which are already interlaced
            t_lms (int): LMS scrambling matrix will be of shape (`t_lms`, `t_max`). Other generating matrix will be of shape (`t_max`, `m_max`).
            _verbose (bool): print randomization details
            replications (int): number of IID randomizations of a pointset
            qmctoolscl_kwargs (dict): keyword arguments for QMCToolsCL to use OpenCL. Defaults to C backend. See https://qmcsoftware.github.io/QMCToolsCL/
        """
        self.parameters = ['dvec','randomize','graycode']
        if alpha>1:
            self.parameters += ["alpha"]
        self.mimics = 'StdUniform'
        self.low_discrepancy = True                      
        self.graycode = graycode
        self.order = "GRAY" if self.graycode else "NATURAL"
        self.replications_gm = 1
        # generating matrices 
        if isinstance(generating_matrices,ndarray):
            if generating_matrices.ndim==2:
                self.z_og = generating_matrices[None,:,:]
            else:
                assert generating_matrices.ndim==3
                self.z_og = generating_matrices
                self.replications_gm = len(self.z_og)
            if d_max is None or t_max is None or m_max is None or msb is None:
                raise ParameterError("d_max, t_max, m_max, and msb must be supplied when generating_matrices is a ndarray")
            self.d_max = d_max
            self.t_max = t_max
            self.m_max = m_max
            self.msb = msb
        elif isinstance(generating_matrices,str):
            parts = generating_matrices.split('.')
            root = dirname(abspath(__file__))+'/generating_matrices/'
            repos = DataSource()
            if "LDData" in generating_matrices and generating_matrices[-4:]!=".npy":
                if repos.exists(generating_matrices):
                    datafile = repos.open(generating_matrices)
                elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices[7:]):
                    datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/LDData/refs/heads/main/dnet/"+generating_matrices[7:])
                elif repos.exists("https://raw.githubusercontent.com/QMCSoftware/"+generating_matrices):
                    datafile = repos.open("https://raw.githubusercontent.com/QMCSoftware/"+generating_matrices)
                else:
                    raise ParameterError("LDData path %s not found"%generating_matrices)
                contents = [line.rstrip('\n').strip() for line in datafile.readlines()]
                self.msb = True
                contents = [line.split("#",1)[0] for line in contents if line[0]!="#"]
                datafile.close()
                assert int(contents[0])==2, "DigitalNetB2 requires base=2 " # base 2
                self.d_max = int(contents[1])
                self.m_max = int(log2(int(contents[2])))
                self.t_max = int(contents[3])
                compat_shift = self.t_max-64 if self.t_max>=64 else 0
                if compat_shift>0: warnings.warn("Truncating ints in generating matrix to have 64 bits.")
                self.z_og = array([[int(v)>>compat_shift for v in line.split(' ')] for line in contents[4:]],dtype=uint64)[None,:]
            elif isfile(root+generating_matrices):
                self.z_og = load(root+generating_matrices).astype(uint64)[None,:]
                self.d_max = int(parts[-5])
                self.t_max = int(parts[-4])
                self.m_max = int(parts[-3])
                self.msb = bool(parts[-2].lower()=='msb')
            elif isfile(generating_matrices):
                self.z_og = load(generating_matrices).astype(uint64)[None,:]
                self.d_max = int(parts[-5])
                self.t_max = int(parts[-4])
                self.m_max = int(parts[-3])
                self.msb = bool(parts[-2].lower()=='msb')
            else:
                raise ParameterError("generating_matrices '%s' not found."%generating_matrices)
        else:
            raise ParameterError("invalid input for generating_matrices, see documentation and doctests.")
        super(DigitalNetB2,self).__init__(dimension,seed)
        if not self.msb:
            qmctoolscl.dnb2_gmat_lsb_to_msb(uint64(self.replications_gm),uint64(self.d),uint64(self.m_max),tile(uint64(self.t_max),self.replications_gm),self.z_og,self.z_og,**qmctoolscl_kwargs)
        # randomizations
        self.t_lms = self.t_max if self.t_max>t_lms else t_lms
        assert self.t_max<=64 and self.t_lms<=64, "require t_max <= t_lms <= 64"
        self.alpha = alpha
        self._verbose = _verbose
        self.randomize = str(randomize).upper()
        if self.randomize=="TRUE": self.randomize = "LMS_DS"
        if self.randomize=="OWEN": self.randomize = "NUS"
        if self.randomize=="NONE": self.randomize = "FALSE"
        if self.randomize=="NO": self.randomize = "FALSE"
        assert self.randomize in ["LMS_DS","LMS","DS","NUS","FALSE"]
        self.z = self.z_og[:,self.dvec].copy()
        if self.alpha>1 and self.randomize in ["DS","FALSE"]:
            assert (self.dvec==arange(self.d)).all()
            dtalpha = self.alpha*self.d
            z_ho = empty((self.replications_gm,self.d,self.m_max),dtype=uint64)
            qmctoolscl.dnb2_interlace(uint64(self.replications_gm),uint64(self.d),uint64(self.m_max),uint64(dtalpha),uint64(self.t_max),uint64(self.t_lms),uint64(self.alpha),self.z_og[:,:dtalpha].copy(),z_ho,**qmctoolscl_kwargs)
            self.z = z_ho
        if "LMS" in self.randomize:
            if self.alpha==1:
                z_lms = empty((replications,self.d,self.m_max),dtype=uint64)
                S = qmctoolscl.dnb2_get_linear_scramble_matrix(self.rng,uint64(replications),uint64(self.d),uint64(self.t_max),uint64(self.t_lms),uint64(self._verbose))
                qmctoolscl.dnb2_linear_matrix_scramble(uint64(replications),uint64(self.d),uint64(self.m_max),uint64(self.replications_gm),uint64(self.t_lms),S,self.z,z_lms,**qmctoolscl_kwargs)
                self.z = z_lms 
                self.replications_gm = replications
            else:
                assert (self.dvec==arange(self.d)).all()
                dtalpha = self.alpha*self.d
                z_lms = empty((replications,dtalpha,self.m_max),dtype=uint64)
                S = qmctoolscl.dnb2_get_linear_scramble_matrix(self.rng,uint64(replications),uint64(dtalpha),uint64(self.t_max),uint64(self.t_lms),uint64(self._verbose))
                qmctoolscl.dnb2_linear_matrix_scramble(uint64(replications),uint64(dtalpha),uint64(self.m_max),uint64(self.replications_gm),uint64(self.t_lms),S,self.z_og[:,:dtalpha].copy(),z_lms,**qmctoolscl_kwargs)
                z_lms_ho = empty((replications,self.d,self.m_max),dtype=uint64)
                qmctoolscl.dnb2_interlace(uint64(replications),uint64(self.d),uint64(self.m_max),uint64(dtalpha),uint64(self.t_lms),uint64(self.t_lms),uint64(self.alpha),z_lms,z_lms_ho,**qmctoolscl_kwargs)
                self.z = z_lms_ho
                self.replications_gm = replications
        if "DS" in self.randomize:
            self.rshift = qmctoolscl.random_tbit_uint64s(self.rng,self.t_lms,(replications,self.d))
        if "NUS" in self.randomize:
            if alpha==1:
                new_seeds = self._base_seed.spawn(replications*self.d)
                self.rngs = array([random.Generator(random.SFC64(new_seeds[j])) for j in range(replications*self.d)]).reshape(replications,self.d)
                self.root_nodes = array([qmctoolscl.NUSNode_dnb2() for i in range(replications*self.d)]).reshape(replications,self.d)
            else:
                assert (self.dvec==arange(self.d)).all()
                self.dtalpha = self.alpha*self.d
                new_seeds = self._base_seed.spawn(replications*self.dtalpha)
                self.rngs = array([random.Generator(random.SFC64(new_seeds[j])) for j in range(replications*self.dtalpha)]).reshape(replications,self.dtalpha)
                self.root_nodes = array([qmctoolscl.NUSNode_dnb2() for i in range(replications*self.dtalpha)]).reshape(replications,self.dtalpha)
                self.z = self.z_og[:,:self.dtalpha].copy()
        if self.replications_gm>1: assert replications==self.replications_gm, "if replications_gm>1 require replications = replications_gm"
        self.replications = replications

    def gen_samples(self, n=None, n_min=0, n_max=8, warn=True, return_unrandomized=False, return_binary=False, qmctoolscl_gen_kwargs={"backend":"c"}, qmctoolscl_rand_kwargs={"backend":"c"}, qmctoolscl_convert_kwargs={"backend":"c"}):
        """
        Generate samples

        Args:
            n (int): if n is supplied, generate from n_min=0 to n_max=n samples. 
                Otherwise use the n_min and n_max explicitly supplied as the following 2 arguments
            n_min (int): Starting index of sequence.
            n_max (int): Final index of sequence.
            return_unrandomized (bool): return randomized samples, unrandomized samples or just randomized samples? 
            return_binary (bool): return binary samples as well? 
            qmctoolscl_gen_kwargs,qmctoolscl_rand_kwargs,qmctoolscl_convert_kwargs (dict): keyword arguments for QMCToolsCL to use OpenCL when generating points, performing randomizations, and converting to floats. Defaults to C backend. See https://qmcsoftware.github.io/QMCToolsCL/


        Returns:
            ndarray: replications x (n_max-n_min) x d (dimension) array of samples
        """
        if n:
            n_min = 0
            n_max = n
        if n_min == 0 and warn and self.randomize in ["FALSE","LMS"]:
            warnings.warn("Non-randomized DigitalNetB2 sequence includes the origin",ParameterWarning)
        if n_max > 2**self.m_max:
            raise ParameterError('DigitalNetB2 generating matrices support up to %d samples.'%(2**self.m_max))
        r_x = uint64(self.replications_gm) 
        n = uint64(n_max-n_min)
        d = uint64(self.dtalpha) if self.randomize=="NUS" and self.alpha>1 else uint64(self.d)
        n_start = uint64(n_min)
        mmax = uint64(self.m_max)
        xb = empty((r_x,n,d),dtype=uint64)
        if self.graycode:
            _ = qmctoolscl.dnb2_gen_gray(r_x,n,d,n_start,mmax,self.z,xb,**qmctoolscl_gen_kwargs)
        else: 
            assert (n_min==0 or log2(n_min)%1==0) and (n_max==0 or log2(n_max)%1==0), "DigitalNetB2 in natural order requires n_min and n_max be 0 or powers of 2"
            _ = qmctoolscl.dnb2_gen_natural(r_x,n,d,n_start,mmax,self.z,xb,**qmctoolscl_gen_kwargs)
        if return_unrandomized or self.randomize in ["FALSE","LMS"]:
            tmaxes_new = tile(uint64(self.t_max if self.randomize=="FALSE" and self.alpha==1 else self.t_lms),int(r_x))
            if not return_binary:
                x = empty((r_x,n,d),dtype=float64)
                _ = qmctoolscl.dnb2_integer_to_float(r_x,n,d,tmaxes_new,xb,x,**qmctoolscl_convert_kwargs)
            if r_x==1: 
                if not return_binary:
                    x = x[0]
                else:
                    xb = xb[0]
        r = uint64(self.replications)
        if "NUS" in self.randomize:
            assert return_unrandomized==False
            if self.alpha==1:
                xrb = empty((r,n,d),dtype=uint64)
                tmax = uint64(self.t_max)
                tmax_new = uint64(self.t_lms)
                _ = qmctoolscl.dnb2_nested_uniform_scramble(r,n,d,r_x,tmax,tmax_new,self.rngs,self.root_nodes,xb,xrb)
                if not return_binary:
                    tmaxes_new = tile(uint64(self.t_lms),int(r))
                    xr = empty((r,n,d),dtype=float64)
                    _ = qmctoolscl.dnb2_integer_to_float(r,n,d,tmaxes_new,xrb,xr,**qmctoolscl_convert_kwargs)
            else:
                d = uint64(self.d)
                dtalpha = uint64(self.dtalpha)
                alpha = uint64(self.alpha)
                xrb_lo = empty((r,n,self.dtalpha),dtype=uint64)
                tmax = uint64(self.t_max)
                tmax_new = uint64(self.t_lms)
                _ = qmctoolscl.dnb2_nested_uniform_scramble(r,n,dtalpha,r_x,tmax,tmax_new,self.rngs,self.root_nodes,xb,xrb_lo)
                xrb_t = empty((r,d,n),dtype=uint64)
                xrb_lo_t = moveaxis(xrb_lo,[1,2],[2,1]).copy() 
                qmctoolscl.dnb2_interlace(r,d,n,dtalpha,tmax_new,tmax_new,alpha,xrb_lo_t,xrb_t)
                xrb = moveaxis(xrb_t,[1,2],[2,1]).copy()
                if not return_binary:
                    tmaxes_new = tile(uint64(self.t_lms),int(r))
                    xr = zeros((r,n,d),dtype=float64)
                    qmctoolscl.dnb2_integer_to_float(r,n,d,tmaxes_new,xrb,xr)
            if r==1:
                if not return_binary:
                    xr=xr[0]
                else:
                    xrb = xrb[0]
        if "DS" in self.randomize:
            xrb = empty((r,n,d),dtype=uint64)
            lshifts = tile(uint64((self.t_lms-self.t_max) if "LMS" not in self.randomize else 0),int(r))
            _ = qmctoolscl.dnb2_digital_shift(r,n,d,r_x,lshifts,xb,self.rshift,xrb,**qmctoolscl_rand_kwargs)
            if not return_binary:
                tmaxes_new = tile(uint64(self.t_lms),int(r))
                xr = empty((r,n,d),dtype=float64)
                _ = qmctoolscl.dnb2_integer_to_float(r,n,d,tmaxes_new,xrb,xr,**qmctoolscl_convert_kwargs)
            if r==1:
                if not return_binary:
                    xr=xr[0]
                else:
                    xrb = xrb[0]
        if self.randomize in ["FALSE","LMS"]:
            if return_binary: 
                return xb
            elif return_unrandomized:
                raise ParameterError("return_unrandomized=True only applies when when randomize='SHIFT'.")
            else:
                return x
        else:
            if return_binary: 
                return xrb
            elif return_unrandomized:
                return xr,x 
            else:
                return xr
    
    def pdf(self, x):
        """ pdf of a standard uniform """
        return ones(x.shape[0], dtype=float) 
            
    def _spawn(self, child_seed, dimension):
        return DigitalNetB2(
            dimension=dimension,
            randomize=self.randomize,
            graycode=self.graycode,
            seed=child_seed,
            generating_matrices=self.z_og,
            d_max=self.d_max,
            t_max=self.t_max,
            m_max=self.m_max,
            msb=True, # self.z_og is put into MSB during first initialization 
            t_lms=self.t_lms,
            replications=self.replications)

class Sobol(DigitalNetB2): pass




from ._accumulate_data import AccumulateData
from .mean_var_data import MeanVarData
from .mean_var_data_rep import MeanVarDataRep
from .ld_transform_data import LDTransformData
from .ld_transform_bayes_data import LDTransformBayesData
from .mlmc_data import MLMCData
from .mlqmc_data import MLQMCData
from .mean_var_data_vec import MeanVarDataVec
try: 
    import gpytorch 
    import torch 
    from .pf_gp_ci_data import PFGPCIData
except: pass



import numpy as np 
import scipy.linalg
import itertools

class _GramMatrix(object):
    def __init__(self, kernel_obj, noise, lbeta1s, lbeta2s, lc1s, lc2s):
        self.kernel_obj = kernel_obj
        self.d = self.kernel_obj.d
        self.npt = self.kernel_obj.npt
        self.ckwargs = self.kernel_obj.ckwargs
        self.torchify = self.kernel_obj.torchify
        self.noise = noise
        self.lbeta1s,self.lc1s,self.t1,self.m1 = self._parse_lbetas_lcs(lbeta1s,lc1s)
        self.lbeta2s,self.lc2s,self.t2,self.m2 = self._parse_lbetas_lcs(lbeta2s,lc2s)
        if self.torchify:
            import torch 
            self.transpose_func = lambda x,dims: torch.permute(x,dims)
            self.clone = lambda x: x.clone()
            self.get_ptr = lambda x: x.data_ptr()
        else:
            self.transpose_func = lambda x,dims: x.transpose(*dims)
            self.clone = lambda x: x.copy()
            self.get_ptr = lambda x: x.ctypes.data
        self.lc1s_og = [self.clone(c1s) for c1s in self.lc1s] 
        self.lc2s_og = [self.clone(c2s) for c2s in self.lc2s]
    def eigh(self, m):
        if self.npt==np:
            evals,evecs = np.linalg.eigh(m)
        else:
            import torch 
            evals,evecs = torch.linalg.eigh(m)
        return evals,evecs
    def cho_solve(self, l, b):
        bis1d = b.ndim==1
        if bis1d:
            b = b[:,None]
        if self.npt==np:
            vcs = np.vectorize(lambda l,b: scipy.linalg.cho_solve((l,True),b),signature="(m,m),(m,k)->(m,k)")
            v = vcs(l,b)
        else:
            import torch 
            v = torch.cholesky_solve(b,l,upper=False)
        if bis1d:
            v = v[:,0]
        return v
    def _parse_lbetas_lcs(self, lbetas, lcs):
        if isinstance(lbetas,int): lbetas = [lbetas*self.npt.ones((1,self.d),dtype=int,**self.ckwargs)]
        elif not isinstance(lbetas,list): lbetas = [self.npt.atleast_2d(lbetas)]
        lbetas = [self.npt.atleast_2d(beta1s) for beta1s in lbetas]
        t = len(lbetas)
        m = np.array([len(beta1s) for beta1s in lbetas],dtype=int)
        assert isinstance(lbetas,list) and all(lbetas[tt].shape==(m[tt],self.d) for tt in range(t))
        if np.isscalar(lcs): lcs = [lcs*self.npt.ones(m[tt],dtype=float,**self.ckwargs) for tt in range(t)]
        elif not isinstance(lcs,list): lcs = [lcs]
        lcs = [self.npt.atleast_1d(c1s) for c1s in lcs] 
        assert isinstance(lcs,list) and all(lcs[tt].shape==(m[tt],) for tt in range(t))
        return lbetas,lcs,t,m
    def cholesky(self, mat):
        try:
            l_chol = self.npt.linalg.cholesky(mat)
        except:
            raise Exception("Cholesky not positive definite, try increasing the noise")
        return l_chol
    def _set_invertible_conds(self, invertible_conds):
        self.invertible = all(cond for cond,error_msg in invertible_conds)
        self.invertible_error_msg = "\n\n"+"\n\n".join(error_msg for cond,error_msg in invertible_conds if not cond)
    def _construct_full_gram_matrix(self, x1, x2, t1, t2, lbeta1s, lbeta2s, lc1s, lc2s):
        gms = np.empty((t1,t2),dtype=object) 
        for tt1,tt2 in itertools.product(range(t1),range(t2)):
            gms[tt1,tt2] = self.kernel_obj(x1,x2,lbeta1s[tt1],lbeta2s[tt2],lc1s[tt1],lc2s[tt2])
        gm = self.npt.vstack([self.npt.hstack([gms[tt1,tt2] for tt2 in range(t2)]) for tt1 in range(t1)])
        return gm
    def _convert_x_to__x(self, x):
        return x
    def _convert__x_to_x(self, _x):
        return _x



from ..util import MethodImplementationError, _univ_repr, DimensionError, ParameterError
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from numpy import *


class TrueMeasure(object):
    """ True Measure abstract class. DO NOT INSTANTIATE. """

    def __init__(self):
        prefix = 'A concrete implementation of TrueMeasure must have '
        if not hasattr(self,'domain'):
            raise ParameterError(prefix + 'self.domain, 2xd ndarray of domain lower bounds (first col) and upper bounds (second col)')
        if not hasattr(self,'range'):
            raise ParameterError(prefix + 'self.range, 2xd ndarray of range lower bounds (first col) and upper bounds (second col)')
        if not hasattr(self,'parameters'):
            self.parameters = []
    
    def _parse_sampler(self, sampler):
        self.sub_compatibility_error = False
        if isinstance(sampler,DiscreteDistribution):
            assert sampler.replications==1, "currently only support discrete distribution with one replication"
            self.transform = self # this is the initial transformation, \Psi_0
            self.d = sampler.d # take the dimension from the discrete distribution
            self.discrete_distrib = sampler
            if sampler.mimics == 'StdUniform':
                if not (self.domain==tile([0,1],(self.d,1))).all():
                    raise ParameterError("The True measure's transform should have unit-cube domain.")
            else:
                raise ParameterError('The %s true measure does not support discrete distributions that mimic a %s.'%\
                    (type(self).__name__,sampler.mimics))
        elif isinstance(sampler,TrueMeasure):
            self.transform = sampler # this is a composed transform, \Psi_j for j>0
            self.parameters += ['transform']
            self.d = sampler.d # take the dimension from the sub-sampler (composed transform)
            self.discrete_distrib = self.transform.discrete_distrib
            if (self.domain!=self.transform.range).any():
                self.sub_compatibility_error = True
            if self.transform.sub_compatibility_error:
                raise ParameterError("The sub-transform domain must match the sub-sub-transform range.")
        else:
            raise ParameterError("sampler input should either be a DiscreteDistribution or TrueMeasure")
    
    def gen_samples(self, *args, **kwargs):
        """
        Generate samples from the discrete distribution
        and transform them via the transform method. 

        Args:
            args (tuple): positional arguments to the discrete distributions gen_samples method
            kwargs (dict): keyword arguments to the discrete distributions gen_samples method
        
        Returns: 
            ndarray: n x d matrix of transformed samples
        """
        x = self.discrete_distrib.gen_samples(*args,**kwargs)
        return self._transform_r(x)

    def _transform_r(self, x):
        """
        Complete transformation (recursive).
        Takes into account composed transforms.  

        Args:
            x: n x d matrix of samples mimicking a standard uniform.

        Returns:
            ndarray: n x d matrix of transformed x.  
        """
        if self.sub_compatibility_error:
            raise ParameterError("The transform domain must match the sub-transform range.")
        if self.transform == self: # is \Psi_0
            return self._transform(x)
        else: # is transform \Psi_j for j>0
            xtf = self.transform._transform_r(x)
            return self._transform(xtf)
            
    def _transform(self, x): 
        """ 
        Transformation for this true measure. 

        Args:
            x: n x d matrix of samples mimicking a standard uniform.

        Returns:
            ndarray: n x d matrix of transformed x.  
        """
        raise MethodImplementationError(self,'_transform. Try setting sampler to be in a PDF TrueMeasure to importance sample by.')
        
    def _jacobian_transform_r(self, x):
        """
        Find the complete Jacobian and completely transformed samples (recursive). 
        Takes into account composed transforms. 

        Args:
            x (ndarray): n x d matrix of samples
        
        Returns:
            ndarray: length n vector of transformed samples at locations of x
            ndarray: length n vector of Jacobian values at locations of x

        Note: If transform is T, then the Jacobian of T is 1/lambda(T(x)) for
        """
        if self.sub_compatibility_error:
            raise ParameterError("The transform domain must match the sub-transform range.")
        if self.transform == self: # is \Psi_0
            t = self._transform(x)
            jac = 1/self._weight(t)
            return t,jac
        else: # is transform \Psi_j for j>0
            t_sub,jac_sub  = self.transform._jacobian_transform_r(x)
            t = self._transform(t_sub) # 
            jac = 1/self._weight(t) # |\Psi1/\lambda(\psi(x))
            return t,jac_sub*jac

    def _weight(self, x):
        """
        Non-negative weight function, lambda. 
        This is often a PDF, but is not required to be 
        i.e. Lebesgue weight is always 1, but is not a PDF.

        Args:
            x (ndarray): n x d  matrix of samples
        
        Returns:
            ndarray: length n vector of weights at locations of x
        """ 
        raise MethodImplementationError(self,'weight. Try a different true measure with a _weight method.')  
    
    def spawn(self, s=1, dimensions=None):
        """
        Spawn new instances of the current discrete distribution but with new seeds and dimensions. 
        Developed for multi-level and multi-replication (Q)MC algorithms.
        
        Args:
            s (int): number of spawn
            dimensions (ndarray): length s array of dimension for each spawn. Defaults to current dimension
        
        Return: 
            list: list of TrueMeasures linked to newly spawned DiscreteDistributions
        """
        if (isinstance(dimensions,list) or isinstance(dimensions,ndarray)) and len(dimensions)==s:
            dimensions = array(dimensions)
        elif isscalar(dimensions) and dimensions%1==0:
            dimensions = tile(dimensions,s)
        elif dimensions is None:
            dimensions = tile(self.d,s)
        else:
            raise ParameterError("invalid spawn dimensions, must be None, int, or length s ndarray")
        spawns = [None]*s
        sampler_spawns = [None]*s
        for i in range(s):
            sampler = self.discrete_distrib if self.transform==self else self.transform
            sampler_spawns[i] = sampler.spawn(s=1,dimensions=[dimensions[i]])[0]
            spawns[i] = self._spawn(sampler_spawns[i],dimensions[i])
        return spawns
    
    def _spawn(self, sampler, dimension):
        """
        ABSTRACT METHOD, used by self.spawn
        
        Args:
            sampler (DiscreteDistribution or TrueMeasure): spawn of sampler for this measure
            dimension (ndarray): dimension of new spawn
        
        Return: 
            TrueMeasure: spawn with dimension and sampler
        """
        raise MethodImplementationError(self, '_spawn')  
        
    def __repr__(self):
        return _univ_repr(self, "TrueMeasure", self.parameters)
    




from ._accumulate_data import AccumulateData
from ..integrand._integrand import Integrand
from ..util import CubatureWarning
from numpy import *
import warnings


class LDTransformData(AccumulateData):
    """
    Update and store transformation data based on low-discrepancy sequences. 
    See the stopping criterion that utilize this object for references.
    """

    def __init__(self, m_min, m_max, coefv, fudge, check_cone, ncv, cv_mu, update_beta):
        self.m_min = m_min
        self.m_max = m_max
        self.coefv = coefv
        self.fudge = fudge
        self.check_cone = check_cone
        self.ncv = ncv
        self.cv_mu = cv_mu
        self.update_beta = update_beta
        self.omg_circ = lambda m: 2**(-m)
        self.r_lag = 4
        self.l_star = int(self.m_min-self.r_lag)
        self.omg_hat = lambda m: self.fudge(m)/((1+self.fudge(self.r_lag))*self.omg_circ(self.r_lag))
        self.c_stilde_low = tile(-inf,int(self.m_max-self.l_star+1))
        self.c_stilde_up = tile(inf,int(self.m_max-self.l_star+1))
        self.y_val = zeros(0,dtype=float)
        self.y_cp = zeros(0)
        self.yg_val = zeros((0,self.ncv),dtype=float)
        self.yg_cp = zeros((0,self.ncv))
        super(LDTransformData,self).__init__()

    def update_data(self, m, y_val_next, y_cp_next, yg_val_next, yg_cp_next):
        self.y_val = hstack((self.y_val,y_val_next))
        self.y_cp = hstack((self.y_cp,y_cp_next))
        self.yg_val = vstack((self.yg_val,yg_val_next))
        self.yg_cp = vstack((self.yg_cp,yg_cp_next))
        mllstart = m-self.r_lag-1
        nllstart = int(2**mllstart)
        if m==self.m_min: # first iteration
            n = int(2**m)
            self.y_cp = self.fast_transform(self.y_cp,0,m,m)
            self.kappanumap = arange(1,n+1,dtype=int)
            self.update_kappanumap(m-1,0,m)
            if self.ncv>0:
                self.yg_cp = self.fast_transform(self.yg_cp,0,m,m)
                self.beta_update(mllstart)
                self.kappanumap = arange(1,n+1,dtype=int)
                self.update_kappanumap(m-1,0,m)
        else: # any iteration after the first
            mnext = int(m-1)
            n = int(2**mnext)
            if self.ncv>0:
                self.y_val[-n:] = self.y_val[-n:]-self.yg_val[-n:]@self.beta
                self.y_cp[-n:] = self.y_val[-n:]
            if not self.update_beta: # do not update the beta coefficients
                self.y_cp[-n:] = self.fast_transform(self.y_cp[-n:],0,mnext,mnext)
                self.y_cp = self.fast_transform(self.y_cp,mnext,mnext+1,mnext)
                self.kappanumap = hstack((self.kappanumap,int(2**(m-1))+self.kappanumap))
                self.update_kappanumap(m-1,mllstart,m) 
            else: # update beta
                self.y_cp = self.fast_transform(self.y_cp,0,m,m)
                self.yg_cp = self.fast_transform(self.yg_cp,0,m,m)
                self.beta_update(mllstart)
                self.kappanumap = hstack((self.kappanumap,2**(m-1)+self.kappanumap)).astype(int)
                self.update_kappanumap(m-1,mllstart,m)
        self.muhat = self.y_val.mean()+self.beta@self.cv_mu if self.ncv>0 else self.y_val.mean()
        stilde = sum(abs(self.y_cp[self.kappanumap[nllstart:2*nllstart]-1]))
        self.bounds = self.muhat+array([-1,1])*self.fudge(m)*stilde
        if self.check_cone:
            for l in range(self.l_star,int(m+1)): # Storing the information for the necessary conditions
                c_tmp = self.omg_hat(m-l)*self.omg_circ(m-l)
                c_low = 1./(1+c_tmp)
                c_up = 1./(1-c_tmp)
                const1 = sum(abs(self.y_cp[self.kappanumap[int(2**(l-1)):int(2**l)]-1]))
                idx = int(l-self.l_star)
                self.c_stilde_low[idx] = maximum(self.c_stilde_low[idx],c_low*const1)
                if c_tmp < 1:
                    self.c_stilde_up[idx] = min(self.c_stilde_up[idx],c_up*const1)
            cone_violation = (self.c_stilde_low > self.c_stilde_up).any()
        else:
            cone_violation = False
        return self.muhat,self.bounds[0],self.bounds[1],cone_violation 

    def fast_transform(self, y2tf, mfrom, mto, m):
        for l in range(int(mfrom),int(mto)):
            nl = 2**l
            nmminlm1 = int(ceil(2**(m-l-1)))
            ptind_nl = hstack((tile(True,nl),tile(False,nl)))
            ptind = tile(ptind_nl,nmminlm1)
            coef = self.coefv(nl)
            coefv = tile(coef,nmminlm1)
            evenval = y2tf[ptind]
            oddval = y2tf[~ptind]
            y2tf[ptind] = (evenval+(coefv*oddval.T).T)/2.
            y2tf[~ptind] = (evenval-(coefv*oddval.T).T)/2.
        return y2tf

    def update_kappanumap(self, mfrom, mto, m):
        for l in range(int(mfrom),int(mto),-1):
            nl = 2**l
            oldone = abs(self.y_cp[self.kappanumap[1:int(nl)]-1]) # earlier values of kappa, don't touch first one
            newone = abs(self.y_cp[self.kappanumap[nl+1:2*nl]-1]) # later values of kappa,
            flip = where(newone>oldone)[0]+1 # which in the pair are the larger ones. change to matlab indexing
            if flip.size!=0:
                additive = arange(0,2**m-1,2**(l+1)).reshape((1,-1))
                flipall = (flip.reshape((-1,1))+additive)
                flipall = flipall.flatten('F').astype(int) # flatten column wise
                temp = self.kappanumap[nl+flipall] # then flip
                self.kappanumap[nl+flipall] = self.kappanumap[flipall] # them
                self.kappanumap[flipall] = temp # around
    
    def beta_update(self,mstart):
        kappa_approx = self.kappanumap[int(2**mstart):]-1 # kappa index used for fitting
        x4beta = self.yg_cp[kappa_approx]
        y4beta = self.y_cp[kappa_approx]
        self.beta = linalg.lstsq(x4beta,y4beta,rcond=None)[0]
        self.y_val = self.y_val-self.yg_val@self.beta # get new function values
        self.y_cp = self.y_cp-self.yg_cp@self.beta # redefine function



from ._stopping_criterion import StoppingCriterion
from ..discrete_distribution import DigitalNetB2
from ..integrand.ishigami import Ishigami
from ..true_measure._true_measure import TrueMeasure
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from ..accumulate_data.pf_gp_ci_data import PFGPCIData, _error_udens
from ..util import MaxSamplesWarning
import warnings
import time
from numpy import *
from scipy.stats import norm
import gpytorch 
import torch

class Suggester(object): pass

class PFSampleErrorDensityAR(Suggester):
    def __init__(self, verbose=False):
        self.verbose = verbose
        super(PFSampleErrorDensityAR,self).__init__()
    def suggest(self, n, d, gp, rng, efficiency, pct=.5):
        if self.verbose: print('\tAR sampling with efficiency %.1e, expect %d draws: '%(efficiency,int(n/efficiency)),end='',flush=True)
        x = zeros((0,d),dtype=float)
        self.n_ar_tries_batch = 0
        rem = n
        draws_per_rem = int(ceil(log(pct)/log(1-efficiency)))
        while rem>0:
            newtries = rem*draws_per_rem
            self.n_ar_tries_batch += newtries
            if self.verbose: print('%d, '%self.n_ar_tries_batch,end='',flush=True)
            z = rng.random((newtries,d))
            u = rng.random(newtries)
            udens_z = _error_udens(gp,z)
            x = vstack([x,z[u<=udens_z]])
            rem = max(n-len(x),0)
        if self.verbose: print()
        return x[:n]

class SuggesterSimple(Suggester):
    def __init__(self, sampler):
        self.sampler = sampler
        if isinstance(self.sampler,TrueMeasure): assert (self.sampler.range==[0,1]).all()
        self.n_min = 0
        super(SuggesterSimple,self).__init__()
    def suggest(self, n, d, gp, rng, **kwargs):
        n_max = self.n_min+n
        assert d == self.sampler.d
        try: x = self.sampler.gen_samples(n_min=self.n_min,n_max=n_max)
        except: x = self.sampler.gen_samples(n)
        self.n_min = n_max
        return x

class PFGPCI(StoppingCriterion):
    """
    Probability of failure estimation using adaptive Gaussian Processes (GP) construction and resulting credible intervals. 
    
    >>> pfgpci = PFGPCI(
    ...     integrand = Ishigami(DigitalNetB2(3,seed=17)),
    ...     failure_threshold = 0,
    ...     failure_above_threshold = False, 
    ...     abs_tol = 2.5e-2,
    ...     alpha = 1e-1,
    ...     n_init = 64,
    ...     init_samples = None,
    ...     batch_sampler = PFSampleErrorDensityAR(verbose=False),
    ...     n_batch = 16,
    ...     n_max = 128,
    ...     n_approx = 2**18,
    ...     gpytorch_prior_mean = gpytorch.means.ZeroMean(),
    ...     gpytorch_prior_cov = gpytorch.kernels.ScaleKernel(
    ...         gpytorch.kernels.MaternKernel(nu=2.5,lengthscale_constraint = gpytorch.constraints.Interval(.5,1)),
    ...         outputscale_constraint = gpytorch.constraints.Interval(1e-8,.5)),
    ...     gpytorch_likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint = gpytorch.constraints.Interval(1e-12,1e-8)),
    ...     gpytorch_marginal_log_likelihood_func = lambda likelihood,gpyt_model: gpytorch.mlls.ExactMarginalLogLikelihood(likelihood,gpyt_model),
    ...     torch_optimizer_func = lambda gpyt_model: torch.optim.Adam(gpyt_model.parameters(),lr=0.1),
    ...     gpytorch_train_iter = 100,
    ...     gpytorch_use_gpu = False,
    ...     verbose = False,
    ...     n_ref_approx = 2**22,
    ...     seed_ref_approx = 11)
    >>> solution,data = pfgpci.integrate(seed=7,refit=True)
    >>> data
    PFGPCIData (AccumulateData Object)
        solution        0.158
        error_bound     0.022
        bound_low       0.136
        bound_high      0.180
        n_total         112
        time_integrate  ...
    PFGPCI (StoppingCriterion Object)
    Ishigami (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     -3.142
        upper_bound     3.142
    DigitalNetB2 (DiscreteDistribution Object)
        d               3
        dvec            [0 1 2]
        randomize       LMS_DS
        graycode        0
        entropy         17
        spawn_key       ()
    >>> df = data.get_results_dict()
    """
    def __init__(self, 
        integrand,
        failure_threshold,
        failure_above_threshold,
        abs_tol = 5e-3,
        alpha = 1e-2,
        n_init = 64,
        init_samples = None,
        batch_sampler = PFSampleErrorDensityAR(),
        n_batch = 4,
        n_max = 1000,
        n_approx = 2**20,
        gpytorch_prior_mean = gpytorch.means.ZeroMean(),
        gpytorch_prior_cov = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel(nu = 2.5)),
        gpytorch_likelihood = gpytorch.likelihoods.GaussianLikelihood(noise_constraint = gpytorch.constraints.Interval(1e-12,1e-8)),
        gpytorch_marginal_log_likelihood_func = lambda likelihood,gpyt_model: gpytorch.mlls.ExactMarginalLogLikelihood(likelihood,gpyt_model),
        torch_optimizer_func = lambda gpyt_model: torch.optim.Adam(gpyt_model.parameters(),lr=0.1),
        gpytorch_train_iter = 100,
        gpytorch_use_gpu = False,
        verbose = False,
        n_ref_approx = 2**22,
        seed_ref_approx = None):
        '''
        Args:
            integrand (Integrand): The simulation whose probability of failure is estimated
            failure_threshold (float): Thresholds for failure. 
            failure_above_threshold (bool): Set to True if failure occurs when the simulation exceeds failure_threshold and False otherwise 
            abs_tol (float): The desired maximum distance from the estimate to either end of the confidence interval
            alpha (float): The credible interval is constructed to hold with probability at least 1 - alpha
            n_init (float): Initial number of samples from integrand.discrete_distrib from which to build the first surrogate GP
            init_samples (float): If the simulation has already been run, pass in (x,y) where x are past samples from the discrete distribution and y are corresponding simulation evaluations. 
            batch_sampler (Suggester or DiscreteDistribution): A suggestion scheme for future samples. 
            n_batch (int): The number of samples per batch to draw from batch_sampler. 
            n_max (int): Budget of simulations.
            n_approx (int): Number of points from integrand.discrete_distrib used to approximate estimate and credible interval bounds
            gpytorch_prior_mean (gpytorch.means): prior mean function of the GP
            gpytorch_prior_cov (gpytorch.kernels): Prior covariance kernel of the GP
            gpytorch_likelihood (gpytorch.likelihoods): GP likelihood, require one of gpytorch.likelihoods.{GaussianLikelihood, GaussianLikelihoodWithMissingObs, FixedNoiseGaussianLikelihood}
            gpytorch_marginal_log_likelihood_func (callable): Function taking in the likelihood and gpytorch model and returning a marginal log likelihood from gpytorch.mlls
            torch_optimizer_func (callable): Function taking in the gpytorch model and returning an optimizer from torch.optim
            gpytorch_train_iter (int): Training iterations for the GP in gpytorch
            gpytorch_use_gpu (bool): If True, have gpytorch use a GPU for fitting and trining the GP
            verbose (int): If verbose > 0, print information through the call to integrate()
            n_ref_approx (int): If n_ref_approx > 0, use n_ref_approx points to get a reference QMC approximation of the true solution. 
                Caution: If n_ref_approx > 0, it should be a large int e.g. 2**22, in which case it is only helpful for cheap to evaluate simulations 
            seed_ref_approx (int): Seed for the reference approximation. Only applies when n_ref_approx>0
        '''
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        self.d = self.integrand.d
        self.sampler = self.d
        self.failure_threshold = failure_threshold
        self.failure_above_threshold = failure_above_threshold
        self.abs_tol = abs_tol
        self.alpha = alpha
        self.n_init = n_init
        self.init_samples = init_samples is not None
        if self.init_samples: 
            self.x_init,self.y_init = init_samples
            assert self.x_init.ndim==2 and self.y_init.ndim==1
            assert self.x_init.shape[1]==self.d and len(self.y_init)==len(self.x_init)
            assert self.n_init==len(self.x_init)
            self.ytf_init = self._affine_tf(self.y_init)
        self.batch_sampler = batch_sampler
        self.n_batch = n_batch
        self.n_max = n_max
        assert self.n_max >= self.n_init
        self.n_approx = n_approx
        assert (self.n_approx+self.n_init) <= 2**32
        self.gpytorch_prior_mean = gpytorch_prior_mean
        self.gpytorch_prior_cov = gpytorch_prior_cov
        self.gpytorch_likelihood = gpytorch_likelihood
        self.gpytorch_marginal_log_likelihood_func = gpytorch_marginal_log_likelihood_func
        self.torch_optimizer_func = torch_optimizer_func
        self.gpytorch_train_iter = gpytorch_train_iter
        self.gpytorch_use_gpu = gpytorch_use_gpu
        self.verbose = verbose
        self.approx_true_solution = n_ref_approx>0
        if self.approx_true_solution: 
            x = DigitalNetB2(self.d,graycode=True,seed=seed_ref_approx).gen_samples(n_ref_approx)
            y = self.integrand.f(x).squeeze()
            ytf = self._affine_tf(y)
            self.ref_approx = (ytf>=0).mean(0)
            if self.verbose: print('reference approximation with d=%d: %s'%(self.d,self.ref_approx))
        super(PFGPCI,self).__init__(allowed_levels=["single"], allowed_distribs=[DiscreteDistribution], allow_vectorized_integrals=False)
    def _affine_tf(self, y):
        return y-self.failure_threshold if self.failure_above_threshold else self.failure_threshold-y
    #@profile
    def integrate(self, seed=None, refit=False):
        t0 = time.time()
        dnb2 = DigitalNetB2(self.d,randomize='DS',graycode=True,seed=seed)
        data = PFGPCIData(self,self.integrand,self.true_measure,self.discrete_distrib,dnb2,self.n_approx,self.alpha,refit, 
                self.gpytorch_prior_mean,
                self.gpytorch_prior_cov,
                self.gpytorch_likelihood,
                self.gpytorch_marginal_log_likelihood_func,
                self.torch_optimizer_func,
                self.gpytorch_train_iter,
                self.gpytorch_use_gpu,
                self.verbose, self.approx_true_solution)
        batch_count = 0
        while True:
            if self.verbose: print('batch %d'%batch_count)
            if batch_count==0:
                if self.init_samples: 
                    xdraw,ydraw = self.x_init,self.y_init
                else:
                    xdraw = dnb2.spawn()[0].gen_samples(self.n_init)
                    ydraw = atleast_1d(self.integrand.f(xdraw).squeeze())
            else:
                n_new = min(self.n_batch,self.n_max-sum(self.n_batch))
                xdraw = self.batch_sampler.suggest(n_new,self.d,data.gpyt_model,dnb2.rng,efficiency=2*data.emr[-1])
                ydraw = atleast_1d(self.integrand.f(xdraw).squeeze())
            ydrawtf = self._affine_tf(ydraw)
            data.update_data(batch_count, xdraw, ydrawtf)
            batch_count += 1
            if data.error_bounds[-1] <= self.abs_tol: break
            if sum(data.n_batch)==self.n_max:
                warnings.warn('n_max reached. ',MaxSamplesWarning)
                break        
        data.solution = data.solutions[-1]
        data.error_bound = data.error_bounds[-1]
        data.bound_low = data.ci_low[-1]
        data.bound_high = data.ci_high[-1]
        data.n_total = sum(data.n_batch)
        data.time_integrate = time.time()-t0
        data.n_sum = cumsum(data.n_batch)
        data.n_batch = array(data.n_batch)
        data.error_bounds = array(data.error_bounds)
        data.ci_low = array(data.ci_low)
        data.ci_high = array(data.ci_high)
        data.solutions = array(data.solutions)
        if self.approx_true_solution:
            data.solutions_ref = tile(self.ref_approx,len(data.n_batch))
            data.error_ref = abs(data.solutions-data.solutions_ref)
            data.in_ci = (data.ci_low<=data.solutions_ref)*(data.solutions_ref<=data.ci_high)
        return data.solution,data



""" Math functions used across various algorithms. """

def _tol_fun(abs_tol, rel_tol, theta, mu, toltype):
    """
    Generalized error tolerance function.

    Args:
        abs_tol (float): absolute error tolerance
        rel_tol (float): relative error tolerance
        theta (float): parameter in 'theta' case
        mu (float): true mean
        toltype (str): different options of tolerance function

    Return:
        float: tolerance as weighted sum of absolute and relative tolerance
    """
    if toltype == 'combine':  # the linear combination of two tolerances
        # theta == 0 --> relative error tolerance
        # theta === 1 --> absolute error tolerance
        tol = theta * abs_tol + (1 - theta) * rel_tol * abs(mu)
    elif toltype == 'max':  # the max case
        tol = max(abs_tol, rel_tol * abs(mu))
    return tol



from ._true_measure import TrueMeasure
from ..util import DimensionError, ParameterError
from ..discrete_distribution import DigitalNetB2
from numpy import *


class Kumaraswamy(TrueMeasure):
    """
    >>> k = Kumaraswamy(DigitalNetB2(2,seed=7),a=[1,2],b=[3,4])
    >>> k.gen_samples(4)
    array([[0.02444474, 0.14161773],
           [0.42937965, 0.53785063],
           [0.11809512, 0.69998807],
           [0.24617103, 0.2723896 ]])
    >>> k
    Kumaraswamy (TrueMeasure Object)
        a               [1 2]
        b               [3 4]
        
    See https://en.wikipedia.org/wiki/Kumaraswamy_distribution
    """

    def __init__(self, sampler, a=2, b=2):
        """
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform 
            a (ndarray): alpha > 0
            b (ndarray): beta > 0
        """
        self.parameters = ['a', 'b']
        self.domain = array([[0,1]])
        self.range = array([[0,1]])
        self._parse_sampler(sampler)
        self.a = a
        self.b = b
        if isscalar(self.a):
            a = tile(self.a,self.d)
        if isscalar(self.b):
            b = tile(self.b,self.d)
        self.alpha = array(a)
        self.beta = array(b)
        if len(self.alpha)!=self.d or len(self.beta)!=self.d:
            raise DimensionError('a and b must be scalar or have length equal to dimension.')
        if not (all(self.alpha>0) and all(self.beta>0)):
            raise ParameterError("Kumaraswamy requires a,b>0.")
        super(Kumaraswamy,self).__init__() 

    def _transform(self, x):
        return (1-(1-x)**(1/self.beta))**(1/self.alpha)
    
    def _weight(self, x):
        return prod( self.alpha*self.beta*x**(self.alpha-1)*(1-x**self.alpha)**(self.beta-1), 1)
    
    def _spawn(self, sampler, dimension):
        if dimension==self.d: # don't do anything if the dimension doesn't change
            spawn = Kumaraswamy(sampler,a=self.alpha,b=self.beta)
        else:
            a = self.alpha[0]
            b = self.beta[0]
            if not (all(self.alpha==a) and all(self.beta==b)):
                raise DimensionError('''
                    In order to spawn a Kumaraswamy measure
                    a must all be the same and 
                    b must all be the same''')
            spawn = Kumaraswamy(sampler,a=a,b=b)
        return spawn


from ._cub_qmc_ld_g import _CubQMCLDG
from ..util import ParameterError
from ..discrete_distribution import DigitalNetB2
from ..true_measure import Gaussian, Uniform
from ..integrand import Keister, CustomFun, BoxIntegral
from numpy import *


class CubQMCNetG(_CubQMCLDG):
    r"""
    Quasi-Monte Carlo method using Sobol' cubature over the
    d-dimensional region to integrate within a specified generalized error
    tolerance with guarantees under Walsh-Fourier coefficients cone decay assumptions.

    >>> k = Keister(DigitalNetB2(2,seed=7))
    >>> sc = CubQMCNetG(k,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> data
    LDTransformData (AccumulateData Object)
        solution        1.808
        comb_bound_low  1.803
        comb_bound_high 1.813
        comb_flags      1
        n_total         2^(10)
        n               2^(10)
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         0.050
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    Keister (Integrand Object)
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    DigitalNetB2 (DiscreteDistribution Object)
        d               2^(1)
        dvec            [0 1]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()
    >>> dd = DigitalNetB2(3,seed=7)
    >>> g1 = CustomFun(Uniform(dd,0,2),lambda t: 10*t[:,0]-5*t[:,1]**2+t[:,2]**3)
    >>> cv1 = CustomFun(Uniform(dd,0,2),lambda t: t[:,0])
    >>> cv2 = CustomFun(Uniform(dd,0,2),lambda t: t[:,1]**2)
    >>> sc = CubQMCNetG(g1,abs_tol=1e-6,check_cone=True,
    ...     control_variates = [cv1,cv2],
    ...     control_variate_means = [1,4/3])
    >>> sol,data = sc.integrate()
    >>> sol
    array([5.33333333])
    >>> exactsol = 16/3
    >>> abs(sol-exactsol)<1e-6
    array([ True])
    >>> dnb2 = DigitalNetB2(3,seed=7)
    >>> f = BoxIntegral(dnb2, s=[-1,1])
    >>> abs_tol = 1e-3
    >>> sc = CubQMCNetG(f, abs_tol=abs_tol)
    >>> solution,data = sc.integrate()
    >>> solution
    array([1.18973505, 0.96056431])
    >>> sol3neg1 = -pi/4-1/2*log(2)+log(5+3*sqrt(3))
    >>> sol31 = sqrt(3)/4+1/2*log(2+sqrt(3))-pi/24
    >>> true_value = array([sol3neg1,sol31])
    >>> (abs(true_value-solution)<abs_tol).all()
    True
    >>> f2 = BoxIntegral(dnb2,s=[3,4])
    >>> sc = CubQMCNetG(f2,control_variates=f,control_variate_means=true_value,update_beta=True)
    >>> solution,data = sc.integrate()
    >>> solution
    array([1.10119574, 1.26661224])
    >>> data
    LDTransformData (AccumulateData Object)
        solution        [1.101 1.267]
        comb_bound_low  [1.099 1.262]
        comb_bound_high [1.104 1.271]
        comb_flags      [ True  True]
        n_total         2^(10)
        n               [1024. 1024.]
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         0.010
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
        cv              BoxIntegral (Integrand Object)
                           s               [-1  1]
        cv_mu           [1.19  0.961]
        update_beta     1
    BoxIntegral (Integrand Object)
        s               [3 4]
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     1
    DigitalNetB2 (DiscreteDistribution Object)
        d               3
        dvec            [0 1 2]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()
    >>> cf = CustomFun(
    ...     true_measure = Uniform(DigitalNetB2(6,seed=7)),
    ...     g = lambda x,compute_flags=None: (2*arange(1,7)*x).reshape(-1,2,3),
    ...     dimension_indv = (2,3))
    >>> sol,data = CubQMCNetG(cf,abs_tol=1e-6).integrate()
    >>> data
    LDTransformData (AccumulateData Object)
        solution        [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_bound_low  [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_bound_high [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_flags      [[ True  True  True]
                        [ True  True  True]]
        n_total         2^(13)
        n               [[4096. 4096. 8192.]
                        [2048. 2048. 4096.]]
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         1.00e-06
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    CustomFun (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     1
    DigitalNetB2 (DiscreteDistribution Object)
        d               6
        dvec            [0 1 2 3 4 5]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()

    Original Implementation:

        https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubSobol_g.m

    References:

        [1] Fred J. Hickernell and Lluis Antoni Jimenez Rugama, 
        Reliable adaptive cubature using digital sequences, 2014. 
        Submitted for publication: arXiv:1410.8615.
        
        [2] Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou, 
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019. 
        Available from http://gailgithub.github.io/GAIL_Dev/

    Guarantee:
        This algorithm computes the integral of real valued functions in $[0,1]^d$
        with a prescribed generalized error tolerance. The Fourier coefficients
        of the integrand are assumed to be absolutely convergent. If the
        algorithm terminates without warning messages, the output is given with
        guarantees under the assumption that the integrand lies inside a cone of
        functions. The guarantee is based on the decay rate of the Fourier
        coefficients. For integration over domains other than $[0,1]^d$, this cone
        condition applies to $f \circ \psi$ (the composition of the
        functions) where $\psi$ is the transformation function for $[0,1]^d$ to
        the desired region. For more details on how the cone is defined, please
        refer to the references below.
    """

    def __init__(self, integrand, abs_tol=1e-2, rel_tol=0., n_init=2.**10, n_max=2.**35,
        fudge=lambda m: 5.*2.**(-m), check_cone=False, 
        control_variates=[], control_variate_means=[], update_beta=False,
        error_fun = lambda sv,abs_tol,rel_tol: maximum(abs_tol,abs(sv)*rel_tol)):

        """
        Args:
            integrand (Integrand): an instance of Integrand
            abs_tol (ndarray): absolute error tolerance
            rel_tol (ndarray): relative error tolerance
            n_init (int): initial number of samples
            n_max (int): maximum number of samples
            fudge (function): positive function multiplying the finite
                              sum of Fast Fourier coefficients specified 
                              in the cone of functions
            check_cone (boolean): check if the function falls in the cone
            control_variates (list): list of integrand objects to be used as control variates. 
                Control variates are currently only compatible with single level problems. 
                The same discrete distribution instance must be used for the integrand and each of the control variates. 
            control_variate_means (list): list of means for each control variate
            update_beta (bool): update control variate beta coefficients at each iteration
            error_fun: function taking in the approximate solution vector, 
                absolute tolerance, and relative tolerance which returns the approximate error. 
                Default indicates integration until either absolute OR relative tolerance is satisfied.
        """
        super(CubQMCNetG,self).__init__(integrand,abs_tol,rel_tol,n_init,n_max,fudge,
            check_cone,control_variates,control_variate_means,update_beta,
            ptransform = 'none',
            coefv = lambda nl: ones(nl,dtype=float), 
            allowed_levels = ['single'],
            allowed_distribs = [DigitalNetB2],
            cast_complex = False,
            error_fun = error_fun)
        if (not self.discrete_distrib.randomize) or self.discrete_distrib.graycode:
            raise ParameterError("CubSobol_g requires distribution to have randomize=True and graycode=False.")


class CubQMCSobolG(CubQMCNetG): pass


from ._true_measure import TrueMeasure
from .brownian_motion import BrownianMotion
from .gaussian import Gaussian, Normal
from .lebesgue import Lebesgue
from .uniform import Uniform
from .kumaraswamy import Kumaraswamy
from .bernoulli_cont import BernoulliCont
from .johnsons_su import JohnsonsSU
from .scipy_wrapper import SciPyWrapper


from ..util import MethodImplementationError, _univ_repr, ParameterError
from ..true_measure._true_measure import TrueMeasure
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from numpy import *
import os
from multiprocessing import get_context
from multiprocessing.pool import ThreadPool
from itertools import repeat


class Integrand(object):
    """ Integrand abstract class. DO NOT INSTANTIATE. """

    def __init__(self, dimension_indv, dimension_comb, parallel, threadpool=False):
        """
        Args:
            dimension_indv (tuple): individual solution shape.
            dimension_comb (tuple): combined solution shape. 
            parallel (int): If parallel is False, 0, or 1: function evaluation is done in serial fashion.
                Otherwise, parallel specifies the number of processes used by 
                multiprocessing.Pool or multiprocessing.pool.ThreadPool.
                Passing parallel=True sets processes = os.cpu_count().
            threadpool (bool): When parallel > 1, 
                if threadpool = True then use multiprocessing.pool.ThreadPool 
                else use multiprocessing.Pool. 
        """
        prefix = 'A concrete implementation of Integrand must have '
        self.d = self.true_measure.d
        self.d_indv = (dimension_indv,) if isinstance(dimension_indv,int) else tuple(dimension_indv)
        self.d_comb = (dimension_comb,) if isinstance(dimension_comb,int) else tuple(dimension_comb)
        cpus = os.cpu_count()
        self.parallel = cpus if parallel is True else int(parallel)
        self.parallel = 0 if self.parallel==1 else self.parallel
        self.threadpool = threadpool
        if not (hasattr(self, 'sampler') and isinstance(self.sampler,(TrueMeasure,DiscreteDistribution))):
            raise ParameterError(prefix + 'self.sampler, a TrueMeasure or DiscreteDistribution instance')
        if not (hasattr(self, 'true_measure') and isinstance(self.true_measure,TrueMeasure)):
            raise ParameterError(prefix + 'self.true_measure, a TrueMeasure instance')
        if not hasattr(self,'parameters'):
            self.parameters = []
        if not hasattr(self,'leveltype'):
            self.leveltype = 'single'
        if not hasattr(self,'max_level'):
            self.max_level = inf
        if not hasattr(self,'discrete_distrib'):
            self.discrete_distrib = self.true_measure.discrete_distrib
        if self.true_measure.transform!=self.true_measure and \
           not (self.true_measure.range==self.true_measure.transform.range).all():
            raise ParameterError("The range of the composed transform is not compatible with this true measure")
        self.EPS = finfo(float).eps

    def g(self, t, compute_flags=None, *args, **kwargs):
        """
        ABSTRACT METHOD for original integrand to be integrated.

        Args:
            t (ndarray): n x d array of samples to be input into original integrand.
            compute_flags (ndarray): outputs that require computation. 
                For example, if the vector function has 3 outputs and compute_flags = [False, True, False], 
                then the function is only required to compute the second output and may leave the remaining outputs as e.g. 0. 
                The False outputs will not be used in the computation since those integrals have been sufficiently approximated.

        Return:
            ndarray: n vector of function evaluations
        """
        raise MethodImplementationError(self, 'g')

    def f(self, x, periodization_transform='NONE', compute_flags=None, *args, **kwargs):
        """
        Evaluate transformed integrand based on true measures and discrete distribution

        Args:
            x (ndarray): n x d array of samples from a discrete distribution
            periodization_transform (str): periodization transform
            compute_flags (ndarray): outputs that require computation. 
                For example, if the vector function has 3 outputs and compute_flags = [False, True, False], 
                then the function is only required to compute the second output and may leave the remaining outputs as e.g. 0. 
                The False outputs will not be used in the computation since those integrals have been sufficiently approximated.
            *args: other ordered args to g
            **kwargs (dict): other keyword args to g

        Return:
            ndarray: length n vector of function evaluations
        """
        periodization_transform = 'NONE' if periodization_transform is None else periodization_transform.upper()
        compute_flags = tile(1,self.d_indv) if compute_flags is None else atleast_1d(compute_flags)
        n,d = x.shape
        # parameter checks
        if self.discrete_distrib.mimics != 'StdUniform' and periodization_transform!='NONE':
            raise ParameterError('''
                Applying a periodization transform currently requires a discrete distribution 
                that mimics a standard uniform measure.''')
        # periodization transform
        if periodization_transform == 'NONE':
            xp = x
            wp = ones(n,dtype=float)
        elif periodization_transform == 'BAKER': # Baker's transform
            xp = 1 - 2 * abs(x - 1 / 2)
            wp = ones(n,dtype=float)
        elif periodization_transform == 'C0': # C^0 transform
            xp = 3 * x ** 2 - 2 * x ** 3
            wp = prod(6 * x * (1 - x), 1)
        elif periodization_transform == 'C1': # C^1 transform
            xp = x ** 3 * (10 - 15 * x + 6 * x ** 2)
            wp = prod(30 * x ** 2 * (1 - x) ** 2, 1)
        elif periodization_transform == 'C1SIN': # Sidi C^1 transform
            xp = x - sin(2 * pi * x) / (2 * pi)
            wp = prod(2 * sin(pi * x) ** 2, 1)
        elif periodization_transform == 'C2SIN': # Sidi C^2 transform
            xp = (8 - 9 * cos(pi * x) + cos(3 * pi * x)) / 16 # psi3
            wp = prod( (9 * sin(pi * x) * pi - sin(3 * pi * x) * 3 * pi) / 16 , 1) # psi3_1
        elif periodization_transform == 'C3SIN': # Sidi C^3 transform
            xp = (12 * pi * x - 8 * sin(2 * pi * x) + sin(4 * pi * x)) / (12 * pi) # psi4
            wp = prod( (12 * pi - 8 * cos(2 * pi * x) * 2 * pi + sin(4 * pi * x) * 4 * pi) / (12 * pi), 1) # psi4_1
        else:
            raise ParameterError("The %s periodization transform is not implemented"%periodization_transform)
        wp = wp.reshape(n)
        if periodization_transform in ['C1','C1SIN','C2SIN','C3SIN']:
            xp[xp<=0] = self.EPS
            xp[xp>=1] = 1-self.EPS
        # function evaluation with chain rule
        y = empty((n,)+self.d_indv,dtype=float)
        if self.true_measure == self.true_measure.transform:
            # jacobian*weight/pdf will cancel so f(x) = g(\Psi(x))
            xtf = self.true_measure._transform(xp) # get transformed samples, equivalent to self.true_measure._transform_r(x)
            y[:] = self._g(xtf,compute_flags,*args,**kwargs)
        else: # using importance sampling --> need to compute pdf, jacobian(s), and weight explicitly
            pdf = self.discrete_distrib.pdf(xp).reshape(n) # pdf of samples
            xtf,jacobians = self.true_measure.transform._jacobian_transform_r(xp) # compute recursive transform+jacobian
            jacobians = jacobians.reshape(n)
            weight = self.true_measure._weight(xtf).reshape(n) # weight based on the true measure
            gvals = self._g(xtf,compute_flags,*args,**kwargs)
            for i in range(n): y[i] = gvals[i]*weight[i]/pdf[i]*jacobians[i]
        # account for periodization weight
        y = (y.T*wp).T
        return y

    def _g(self,t,compute_flags,*args,**kwargs):
        n = len(t)
        kwargs['compute_flags'] = compute_flags
        if self.parallel:
            pool = get_context(method='fork').Pool(processes=self.parallel) if self.threadpool==False else ThreadPool(processes=self.parallel) 
            y = pool.starmap(self._g2,zip(t,repeat((args,kwargs))))
            pool.close()
            y = concatenate(y,dtype=float)
        else:
            y = self._g2(t,comb_args=(args,kwargs))
        y = y.reshape((n,)+self.d_indv)
        return y

    def _g2(self,t,comb_args=((),{})):
        args = comb_args[0]
        kwargs = comb_args[1]
        t = atleast_2d(t)
        if self.d_indv==(1,):
            kwargs = dict(kwargs)
            del kwargs['compute_flags']
        y = self.g(t,*args,**kwargs)
        return y

    def bound_fun(self, bound_low, bound_high):
        """
        Compute the bounds on the combined function based on bounds for the
        individual functions. Defaults to the identity where we essentially
        do not combine integrands, but instead integrate each function
        individually.

        Args:
            bound_low (ndarray): length Integrand.d_indv lower error bound
            bound_high (ndarray): length Integrand.d_indv upper error bound

        Return:
            (tuple) containing

            - (ndarray): lower bound on function combining estimates
            - (ndarray): upper bound on function combining estimates
            - (ndarray): bool flags to override sufficient combined integrand estimation, e.g., when approximating a ratio of integrals, if the denominator's bounds straddle 0, then returning True here forces ratio to be flagged as insufficiently approximated.
        """
        if self.d_indv!=self.d_comb:
            raise ParameterError('''
                Set bound_fun explicitly. 
                The default bound_fun is the identity map. 
                Since individual solution dimension d_indv != combined solution dimension d_comb, 
                QMCPy cannot infer a reasonable bound function.''')
        return bound_low,bound_high

    def dependency(self, comb_flags):
        """
        Takes a vector of indicators of weather of not
        the error bound is satisfied for combined integrands and which returns flags for individual integrands.
        For example, if we are taking the ratio of 2 individual integrands, then getting flag_comb=True means the ratio
        has not been approximated to within the tolerance, so the dependency function should return [True,True]
        indicating that both the numerator and denominator integrands need to be better approximated.
        Args:
            comb_flags (bool ndarray): flags indicating weather the combined integrals are insufficiently approximated

        Return:
            (bool ndarray): length (Integrand.d_indv) flags for individual integrands
        """
        return comb_flags if self.d_indv==self.d_comb else  tile((comb_flags==False).any(),self.d_indv)

    def _dimension_at_level(self, level):
        """
        ABSTRACT METHOD to return the dimension of samples to generate at level l.
        This method only needs to be implemented for multi-level integrands where
        the dimension changes depending on the level.

        Will default to return the current dimension (not using multilevel methods).
        Overwrite this method for multilevel integrands

        Args:
            level (int): level at which to return the dimension

        Return:
            int: dimension at input level
        """
        if self.leveltype!='single':
            raise MethodImplementationError(self, '_dimension_at_level')
        return self.d

    def spawn(self, levels):
        """
        Spawn new instances of the current integrand at the specified levels.

        Args:
            levels (ndarray): array of levels at which to spawn new integrands

        Return:
            list: list of Integrands linked to newly spawned TrueMeasures and DiscreteDistributions
        """
        levels = array([levels]) if isscalar(levels) else array(levels)
        if (levels>self.max_level).any():
            raise ParameterError("requested spawn level exceeds max level")
        n_levels = len(levels)
        spawns = [None]*n_levels
        for l in range(n_levels):
            level = levels[l]
            new_dim = self._dimension_at_level(level)
            tm_spawn = self.sampler.spawn(s=1,dimensions=[new_dim])[0]
            spawns[l] = self._spawn(level,tm_spawn)
        return spawns

    def _spawn(self, level, tm_spawn):
        """
        ABSTRACT METHOD, used by self.spawn

        Args:
            level (numpy.random.SeedSequence): level at which to spawn new instance
            tm_spawn (TrueMeasure): true measure spawn to use as new sampler

        Return:
            Integrand: spawn at this level
        """
        raise MethodImplementationError(self, '_spawn')

    def __repr__(self):
        return _univ_repr(self, "Integrand", self.parameters)



from ._cub_qmc_ld_g import _CubQMCLDG
from ..discrete_distribution import Lattice
from ..true_measure import Gaussian, Uniform
from ..integrand import Keister, BoxIntegral, CustomFun
from ..util import ParameterError
from numpy import *


class CubQMCLatticeG(_CubQMCLDG):
    r"""
    Stopping Criterion quasi-Monte Carlo method using rank-1 Lattices cubature over
    a d-dimensional region to integrate within a specified generalized error
    tolerance with guarantees under Fourier coefficients cone decay assumptions.
    
    >>> k = Keister(Lattice(2,seed=7))
    >>> sc = CubQMCLatticeG(k,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> data
    LDTransformData (AccumulateData Object)
        solution        1.810
        comb_bound_low  1.806
        comb_bound_high 1.815
        comb_flags      1
        n_total         2^(10)
        n               2^(10)
        time_integrate  ...
    CubQMCLatticeG (StoppingCriterion Object)
        abs_tol         0.050
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    Keister (Integrand Object)
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    Lattice (DiscreteDistribution Object)
        d               2^(1)
        dvec            [0 1]
        randomize       SHIFT
        order           NATURAL
        gen_vec         [     1 182667]
        entropy         7
        spawn_key       ()
    >>> f = BoxIntegral(Lattice(3,seed=7), s=[-1,1])
    >>> abs_tol = 1e-3
    >>> sc = CubQMCLatticeG(f, abs_tol=abs_tol)
    >>> solution,data = sc.integrate()
    >>> solution
    array([1.1894253 , 0.96059384])
    >>> sol3neg1 = -pi/4-1/2*log(2)+log(5+3*sqrt(3))
    >>> sol31 = sqrt(3)/4+1/2*log(2+sqrt(3))-pi/24
    >>> true_value = array([sol3neg1,sol31])
    >>> assert (abs(true_value-solution)<abs_tol).all()
    >>> cf = CustomFun(
    ...     true_measure = Uniform(Lattice(6,seed=7)),
    ...     g = lambda x,compute_flags=None: (2*arange(1,7)*x).reshape(-1,2,3),
    ...     dimension_indv = (2,3))
    >>> sol,data = CubQMCLatticeG(cf,abs_tol=1e-6).integrate()
    >>> data
    LDTransformData (AccumulateData Object)
        solution        [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_bound_low  [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_bound_high [[1. 2. 3.]
                        [4. 5. 6.]]
        comb_flags      [[ True  True  True]
                        [ True  True  True]]
        n_total         2^(15)
        n               [[ 8192. 16384. 16384.]
                        [16384. 32768. 32768.]]
        time_integrate  ...
    CubQMCLatticeG (StoppingCriterion Object)
        abs_tol         1.00e-06
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    CustomFun (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     0
        upper_bound     1
    Lattice (DiscreteDistribution Object)
        d               6
        dvec            [0 1 2 3 4 5]
        randomize       SHIFT
        order           NATURAL
        gen_vec         [     1 182667 213731 255351  96013 116671]
        entropy         7
        spawn_key       ()
    
    Original Implementation:

        https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/cubLattice_g.m

    References:

        [1] Lluis Antoni Jimenez Rugama and Fred J. Hickernell, 
        "Adaptive multidimensional integration based on rank-1 lattices," 
        Monte Carlo and Quasi-Monte Carlo Methods: MCQMC, Leuven, Belgium, 
        April 2014 (R. Cools and D. Nuyens, eds.), Springer Proceedings in Mathematics 
        and Statistics, vol. 163, Springer-Verlag, Berlin, 2016, arXiv:1411.1966, pp. 407-422.
        
        [2] Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou, 
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019. 
        Available from http://gailgithub.github.io/GAIL_Dev/

    Guarantee:
        This algorithm computes the integral of real valued functions in $[0,1]^d$
        with a prescribed generalized error tolerance. The Fourier coefficients
        of the integrand are assumed to be absolutely convergent. If the
        algorithm terminates without warning messages, the output is given with
        guarantees under the assumption that the integrand lies inside a cone of
        functions. The guarantee is based on the decay rate of the Fourier
        coefficients. For integration over domains other than $[0,1]^d$, this cone
        condition applies to $f \circ \psi$ (the composition of the
        functions) where $\psi$ is the transformation function for $[0,1]^d$ to
        the desired region. For more details on how the cone is defined, please
        refer to the references below.
    """

    def __init__(self, integrand, abs_tol=1e-2, rel_tol=0., n_init=2.**10, n_max=2.**35,
        fudge=lambda m: 5.*2.**(-m), check_cone=False, ptransform='Baker',
        error_fun = lambda sv,abs_tol,rel_tol: maximum(abs_tol,abs(sv)*rel_tol)):
        """
        Args:
            integrand (Integrand): an instance of Integrand
            abs_tol (ndarray): absolute error tolerance
            rel_tol (ndarray): relative error tolerance
            n_init (int): initial number of samples
            n_max (int): maximum number of samples
            fudge (function): positive function multiplying the finite
                              sum of Fast Fourier coefficients specified 
                              in the cone of functions
            check_cone (boolean): check if the function falls in the cone
            error_fun: function taking in the approximate solution vector, 
                absolute tolerance, and relative tolerance which returns the approximate error. 
                Default indicates integration until either absolute OR relative tolerance is satisfied.
        """
        super(CubQMCLatticeG,self).__init__(integrand,abs_tol,rel_tol,n_init,n_max,fudge,check_cone,
            control_variates = [],
            control_variate_means = [],
            update_beta=False,
            ptransform = ptransform,
            coefv = lambda nl: exp(-2*pi*1j*arange(nl)/(2*nl)), 
            allowed_levels = ['single'],
            allowed_distribs = [Lattice],
            cast_complex = True,
            error_fun = error_fun)
        if not self.discrete_distrib.randomize:
            raise ParameterError("CubLattice_g requires distribution to have randomize=True")
        if self.discrete_distrib.order != 'NATURAL':
            raise ParameterError("CubLattice_g requires Lattice with 'NATURAL' order")



from ._stopping_criterion import StoppingCriterion
from ..accumulate_data import MeanVarData
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from ..discrete_distribution._discrete_distribution import IID
from ..integrand import Keister, CustomFun
from ..true_measure import Gaussian, Uniform
from ..discrete_distribution import IIDStdUniform
from ..util import _tol_fun, MaxSamplesWarning
from numpy import *
from scipy.optimize import fsolve
from scipy.stats import norm
from time import time
import warnings


class CubMCG(StoppingCriterion):
    r"""
    Stopping criterion with guaranteed accuracy.

    >>> k = Keister(IIDStdUniform(2,seed=7))
    >>> sc = CubMCG(k,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> data
    MeanVarData (AccumulateData Object)
        solution        1.807
        error_bound     0.050
        n_total         15256
        n               14232
        levels          1
        time_integrate  ...
    CubMCG (StoppingCriterion Object)
        abs_tol         0.050
        rel_tol         0
        n_init          2^(10)
        n_max           10000000000
        inflate         1.200
        alpha           0.010
    Keister (Integrand Object)
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      2^(-1)
        decomp_type     PCA
    IIDStdUniform (DiscreteDistribution Object)
        d               2^(1)
        entropy         7
        spawn_key       ()
    >>> dd = IIDStdUniform(1,seed=7)
    >>> k = Keister(dd)
    >>> cv1 = CustomFun(Uniform(dd),lambda x: sin(pi*x).sum(1))
    >>> cv1mean = 2/pi
    >>> cv2 = CustomFun(Uniform(dd),lambda x: (-3*(x-.5)**2+1).sum(1))
    >>> cv2mean = 3/4
    >>> sc1 = CubMCG(k,abs_tol=.05,control_variates=[cv1,cv2],control_variate_means=[cv1mean,cv2mean])
    >>> sol,data = sc1.integrate()
    >>> sol
    1.384...

    Original Implementation:

        https://github.com/GailGithub/GAIL_Dev/blob/master/Algorithms/IntegrationExpectation/meanMC_g.m

    References:

        [1] Fred J. Hickernell, Lan Jiang, Yuewei Liu, and Art B. Owen, "Guaranteed
        conservative fixed width confidence intervals via Monte Carlo
        sampling," Monte Carlo and Quasi-Monte Carlo Methods 2012 (J. Dick, F.
        Y. Kuo, G. W. Peters, and I. H. Sloan, eds.), pp. 105-128,
        Springer-Verlag, Berlin, 2014. DOI: 10.1007/978-3-642-41095-6_5

        [2] Sou-Cheng T. Choi, Yuhan Ding, Fred J. Hickernell, Lan Jiang, Lluis Antoni Jimenez Rugama,
        Da Li, Jagadeeswaran Rathinavel, Xin Tong, Kan Zhang, Yizhi Zhang, and Xuan Zhou,
        GAIL: Guaranteed Automatic Integration Library (Version 2.3) [MATLAB Software], 2019.
        Available from http://gailgithub.github.io/GAIL_Dev/

    Guarantee:
        This algorithm attempts to calculate the mean, mu, of a random variable
        to a prescribed error tolerance, _tol_fun:= max(abstol,reltol*|mu|), with
        guaranteed confidence level 1-alpha. If the algorithm terminates without
        showing any warning messages and provides an answer tmu, then the follow
        inequality would be satisfied: $\P(| mu - tmu | \\le tol\_fun) \\ge 1 - \\alpha$.
    """

    def __init__(self, integrand, abs_tol=1e-2, rel_tol=0., n_init=1024., n_max=1e10,
                 inflate=1.2, alpha=0.01, control_variates=[], control_variate_means=[]):
        """
        Args:
            integrand (Integrand): an instance of Integrand
            inflate: inflation factor when estimating variance
            alpha: significance level for confidence interval
            abs_tol: absolute error tolerance
            rel_tol: relative error tolerance
            n_init: initial number of samples
            n_max: maximum number of samples
            control_variates (list): list of integrand objects to be used as control variates.
                Control variates are currently only compatible with single level problems. 
                The same discrete distribution instance must be used for the integrand and each of the control variates. 
            control_variate_means (list): list of means for each control variate
        """
        self.parameters = ['abs_tol','rel_tol','n_init','n_max','inflate','alpha']
        # Set Attributes
        self.abs_tol = float(abs_tol)
        self.rel_tol = float(rel_tol)
        self.n_init = float(n_init)
        self.n_max = float(n_max)
        self.alpha = float(alpha)
        self.inflate = float(inflate)
        self.alpha_sigma = float(self.alpha) / 2.  # the uncertainty for variance estimation
        self.kurtmax = (n_init - 3) / (n_init - 1) + \
            (self.alpha_sigma * n_init) / (1 - self.alpha_sigma) * \
            (1 - 1. / self.inflate**2)**2
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        self.cv = control_variates
        self.cv_mu = control_variate_means
        # Verify Compliant Construction
        allowed_levels = ['single']
        allowed_distribs = [IID]
        allow_vectorized_integrals = False
        super(CubMCG,self).__init__(allowed_levels, allowed_distribs, allow_vectorized_integrals)

    def integrate(self):
        """ See abstract method. """
        # Construct AccumulateData Object to House Integration data
        self.data = MeanVarData(self, self.integrand, self.true_measure, self.discrete_distrib, 
            self.n_init, self.cv, self.cv_mu)  # house integration data
        t_start = time()
        # Pilot Sample
        self.data.update_data()
        self.sigma_up = self.inflate * self.data.sighat
        if self.rel_tol == 0:
            self.alpha_mu = 1 - (1 - self.alpha) / (1 - self.alpha_sigma)
            toloversig = self.abs_tol / self.sigma_up
            # absolute error tolerance over sigma
            n, self.data.error_bound = \
                self._nchebe(toloversig, self.alpha_mu, self.kurtmax, self.n_max, self.sigma_up)
            self.data.n[:] = n
            if self.data.n_total + self.data.n > self.n_max:
                # cannot generate this many new samples
                n_low = int(self.n_max-self.data.n_total)
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples would exceeds n_max = %d.
                Will instead generate %d samples to meet n_max total samples. 
                Note that error tolerances may no longer be satisfied""" \
                % (int(self.data.n_total), int(self.data.n), int(self.n_max), n_low)
                warnings.warn(warning_s, MaxSamplesWarning)
                self.data.n[:] = n_low
            self.data.update_data()
        else: # self.rel_tol > 0
            alphai = (self.alpha-self.alpha_sigma)/(2*(1-self.alpha_sigma)) # uncertainty to do iteration
            eps1 = self._ncbinv(1e4,alphai,self.kurtmax)
            self.data.error_bound = self.sigma_up*eps1
            tau = 1. # step of the iteration
            self.data.n[:] = 1e4 # default initial sample size
            while True:
                if self.data.n_total + self.data.n > self.n_max:
                    # cannot generate this many new samples
                    n_low = int(self.n_max-self.data.n_total)
                    warning_s = """
                    Already generated %d samples.
                    Trying to generate %d new samples would exceeds n_max = %d.
                    Will instead generate %d samples to meet n_max total samples. 
                    Note that error tolerances may no longer be satisfied""" \
                    % (int(self.data.n_total), int(self.data.n), int(self.n_max), n_low)
                    warnings.warn(warning_s, MaxSamplesWarning)
                    self.data.n[:] = n_low
                    self.data.update_data()
                    break
                self.data.update_data()
                lb_tol = _tol_fun(self.abs_tol, self.rel_tol, 0., self.data.solution-self.data.error_bound, 'max')
                ub_tol = _tol_fun(self.abs_tol, self.rel_tol, 0., self.data.solution+self.data.error_bound, 'max')
                delta_plus = (lb_tol + ub_tol) / 2.
                if delta_plus >= self.data.error_bound:
                    # stopping criterion met
                    delta_minus = (lb_tol - ub_tol) / 2.
                    self.data.solution += delta_minus # adjust solution a bit
                    break
                else:
                    candidate_tol = max(self.abs_tol,.95*self.rel_tol*abs(self.data.solution))
                    self.data.error_bound = min(self.data.error_bound/2.,candidate_tol)
                    tau += 1
                # update next uncertainty
                toloversig = self.data.error_bound / self.sigma_up
                alphai = 2**tau * (self.alpha - self.alpha_sigma) / (1 - self.alpha_sigma)
                n,_ = self._nchebe(toloversig, alphai, self.kurtmax, self.n_max, self.sigma_up)
                self.data.n[:] = n
        # set confidence interval
        self.data.confid_int = self.data.solution + self.data.error_bound * array([-1, 1])
        self.data.time_integrate = time() - t_start
        return self.data.solution, self.data

    def _nchebe(self, toloversig, alpha, kurtmax, n_budget, sigma_0_up):
        ncheb = ceil(1 / (alpha * toloversig**2))  # sample size by Chebyshev's Inequality
        A = 18.1139
        A1 = 0.3328
        A2 = 0.429  # three constants in Berry-Esseen inequality
        M3upper = kurtmax**(3. / 4)
        # the upper bound on the third moment by Jensen's inequality
        BEfun2 = lambda logsqrtn: \
            (norm.cdf(-exp(logsqrtn) * toloversig)
            + exp(-logsqrtn) * minimum(A1 * (M3upper + A2),
            A * M3upper / (1 + (exp(logsqrtn) * toloversig)**3))
            - alpha / 2.)
        # Berry-Esseen function, whose solution is the sample size needed
        logsqrtnCLT = log(norm.ppf(1 - alpha / 2) / toloversig)  # sample size by CLT
        nbe = ceil(exp(2 * fsolve(BEfun2, logsqrtnCLT)))
        # calculate Berry-Esseen n by fsolve function (scipy)
        ncb = min(min(ncheb, nbe), n_budget)  # take the min of two sample sizes
        logsqrtn = log(sqrt(ncb))
        BEfun3 = lambda toloversig: \
            (norm.cdf(-exp(logsqrtn) * toloversig)
            + exp(-logsqrtn) * min(A1 * (M3upper + A2),
            A * M3upper / (1 + (exp(logsqrtn) * toloversig)**3))
            - alpha / 2.)
        err = fsolve(BEfun3, toloversig) * sigma_0_up
        return ncb, err

    def _ncbinv(self, n1, alpha1, kurtmax):
        NCheb_inv = 1/sqrt(n1*alpha1)
        # use Chebyshev inequality
        A = 18.1139
        A1 = 0.3328
        A2 = 0.429 # three constants in Berry-Esseen inequality
        M3upper = kurtmax**(3./4)
        # using Jensen's inequality to bound the third moment
        BEfun = lambda logsqrtb: \
            (norm.cdf(n1*logsqrtb) + min(A1*(M3upper+A2),
            A*M3upper/(1+(sqrt(n1)*logsqrtb)**3))/sqrt(n1)
            - alpha1/2)
        # Berry-Esseen inequality
        logsqrtb_clt = log(sqrt(norm.ppf(1-alpha1/2)/sqrt(n1)))
        # use CLT to get tolerance
        NBE_inv = exp(2*fsolve(BEfun,logsqrtb_clt))
        # use fsolve to get Berry-Esseen tolerance
        eps = min(NCheb_inv,NBE_inv)
        # take the min of Chebyshev and Berry Esseen tolerance
        return eps

    def set_tolerance(self, abs_tol=None, rel_tol=None):
        """
        See abstract method.

        Args:
            abs_tol (float): absolute tolerance. Reset if supplied, ignored if not.
            rel_tol (float): relative tolerance. Reset if supplied, ignored if not.
        """
        if abs_tol != None: self.abs_tol = abs_tol
        if rel_tol != None: self.rel_tol = rel_tol


from ..util import ParameterError, MethodImplementationError, _univ_repr, DimensionError
from numpy import *

class DiscreteDistribution(object):
    """ Discrete Distribution abstract class. DO NOT INSTANTIATE. """

    def __init__(self, dimension, seed):
        """
        Args:
            dimension (int or ndarray): dimension of the generator.
                If an int is passed in, use sequence dimensions [0,...,dimensions-1].
                If a ndarray is passed in, use these dimension indices in the sequence.
                Note that this is not relevant for IID generators.
            seed (int or numpy.random.SeedSequence): seed to create random number generator
        """
        prefix = 'A concrete implementation of DiscreteDistribution must have '
        if not hasattr(self, 'mimics'):
            raise ParameterError(prefix + 'self.mimcs (measure mimicked by the distribution)')
        if not hasattr(self,'low_discrepancy'):
            raise ParameterError(prefix + 'self.low_discrepancy')
        if not hasattr(self,'parameters'):
            self.parameters = []
        if not hasattr(self,'d_max'):
            raise ParameterError(prefix+ 'self.d_max')
        if isinstance(dimension,list) or isinstance(dimension,ndarray):
            self.dvec = array(dimension)
            self.d = len(self.dvec)
        else:
            self.d = dimension
            self.dvec = arange(self.d)
        if any(self.dvec>self.d_max):
            raise ParameterError('dimension greater than max dimension %d'%self.d_max)
        self._base_seed = seed if isinstance(seed,random.SeedSequence) else random.SeedSequence(seed)
        self.entropy = self._base_seed.entropy
        self.spawn_key = self._base_seed.spawn_key
        self.rng = random.Generator(random.SFC64(self._base_seed))

    def gen_samples(self, *args):
        """
        ABSTRACT METHOD to generate samples from this discrete distribution.

        Args:
            args (tuple): tuple of positional argument. See implementations for details

        Returns:
            ndarray: n x d array of samples
        """
        raise MethodImplementationError(self, 'gen_samples')

    def pdf(self, x):
        """ ABSTRACT METHOD to evaluate pdf of distribution the samples mimic at locations of x. """
        raise MethodImplementationError(self, 'pdf')

    
    def spawn(self, s=1, dimensions=None):
        """
        Spawn new instances of the current discrete distribution but with new seeds and dimensions.
        Developed for multi-level and multi-replication (Q)MC algorithms.

        Args:
            s (int): number of spawn
            dimensions (ndarray): length s array of dimension for each spawn. Defaults to current dimension

        Return:
            list: list of DiscreteDistribution instances with new seeds and dimensions
        """
        if (isinstance(dimensions,list) or isinstance(dimensions,ndarray)) and len(dimensions)==s:
            dimensions = array(dimensions)
        elif isscalar(dimensions) and dimensions%1==0:
            dimensions = tile(dimensions,s)
        elif dimensions is None:
            dimensions = tile(self.d,s)
        else:
            raise ParameterError("invalid spawn dimensions, must be None, int, or length s ndarray")
        child_seeds = self._base_seed.spawn(s)
        return [self._spawn(child_seeds[i],dimensions[i]) for i in range(s)]

    def _spawn(self, child_seed, dimension):
        """
        ABSTRACT METHOD, used by self.spawn

        Args:
            child_seeds (numpy.random.SeedSequence): length s array of seeds for each spawn
            dimension (int): length s array of dimensions for each spawn

        Return:
            DiscreteDistribution: spawn with new dimension using child_seed
        """
        raise MethodImplementationError(self, '_spawn')

    def __repr__(self):
        return _univ_repr(self, "DiscreteDistribution", ['d']+self.parameters+['entropy','spawn_key'])

    def __call__(self, *args, **kwargs):
        if len(args)>2 or len(args)==0:
            raise Exception('''
                expecting 1 or 2 arguments:
                    1 argument corresponds to n, the number of samples to generate. In this case n_min=0 and n_max=n for LD sequences
                    2 arguments corresponds to n_min and n_max. Note this is incompatible with IID generators which only expect 1 argument.
                ''')
        if len(args) == 1:
            return self.gen_samples(n=args[0])
        else:
            return self.gen_samples(n_min=args[0],n_max=args[1])

class LD(DiscreteDistribution): pass

class IID(DiscreteDistribution): pass


import numpy as np

class _KernelProd(object):
    def __init__(self, dimension, lengthscales=1., scale=1., torchify=False, device="cpu", alpha=None, requires_grad=False, **kwargs):
        """
        Args:
            dimension (int): dimension of the input
            lengthscales (np.ndarray or torch.Tensor): vector of length scales of length d 
            scale (float): global scale 
            torchify (bool): wheather or not to use PyTorch backend
            alpha (np.ndarray or torch.Tensor): vector of smoothness parameters of length d 
        """
        self.d = dimension
        self.torchify = torchify 
        if self.torchify:
            import torch 
            self.npt = torch
            assert device!="mps", "mps does not (yet) support complex datatypes"
            self.ckwargs = {"device":device}
        else: 
            self.npt = np
            self.ckwargs = {}
        assert np.isscalar(lengthscales) or lengthscales.shape==(self.d,)
        assert np.isscalar(scale) or (scale.ndim==1 and len(scale)==1)
        if np.isscalar(lengthscales):
            lengthscales = lengthscales*self.npt.ones(self.d,dtype=float,**self.ckwargs)
        self.lengthscales = lengthscales
        self.scale = scale
        if alpha is None: alpha = self.DEFAULTALPHA
        assert alpha is np.nan or isinstance(alpha,int) or alpha.shape==(self.d,)
        if isinstance(alpha,int):
            alpha = alpha*self.npt.ones(self.d,dtype=int,**self.ckwargs)
        self.alpha = alpha
        self.requires_grad = requires_grad
        self._my_init(**kwargs)
    def _my_init(self,**kwargs):
        pass
    def eval(self, *args, **kwargs):
        return self.__call__(*args,**kwargs)
    def __call__(self, x1, x2, beta1s=0, beta2s=0, c1s=1., c2s=1., diag_only=False):
        """
        Args:
            x1 (np.ndarray or torch.Tensor): n1 x d array of first inputs to kernel 
            x2 (np.ndarray or torch.Tensor): n2 x d array of second inputs to kernel
            beta1s (np.ndarray or torch.Tensor): m1 x d array of first derivative orders
            beta2s (np.ndarray or torch.Tensor): m2 x d array of first derivative orders
            c1s (np.ndarray or torch.Tensor): length m1 vector of derivative coefficients 
            c2s (np.ndarray or torch.Tensor): length m2 vector of derivative coefficients
            diag_only (bool): only compute the diagonal of the covariance matrix 
        """
        n1,n2 = len(x1),len(x2)
        assert x1.shape==(n1,self.d) and x2.shape==(n2,self.d)
        if isinstance(beta1s,int): beta1s = beta1s*self.npt.ones(self.d,dtype=int,**self.ckwargs)
        if isinstance(beta2s,int): beta2s = beta2s*self.npt.ones(self.d,dtype=int,**self.ckwargs)
        beta1s = self.npt.atleast_2d(beta1s)
        beta2s = self.npt.atleast_2d(beta2s)
        m1 = len(beta1s) 
        m2 = len(beta2s)
        assert beta1s.shape==(m1,self.d) and beta2s.shape==(m2,self.d)
        if np.isscalar(c1s): c1s = c1s*self.npt.ones(m1,dtype=float,**self.ckwargs)
        if np.isscalar(c2s): c2s = c2s*self.npt.ones(m2,dtype=float,**self.ckwargs)
        assert c1s.shape==(m1,) and c2s.shape==(m2,)
        if diag_only:
            assert n1==n2 and m1==m2
        return self.parsed_call(x1,x2,n1,n2,beta1s,beta2s,m1,m2,c1s,c2s,diag_only)



from ._integrand import Integrand
from ..discrete_distribution import DigitalNetB2
from ..stopping_criterion import CubQMCNetG
from ..true_measure import Uniform
from ..util import ParameterError
from numpy import *
import os

class UMBridgeWrapper(Integrand):
    """
    UM-Bridge Model Wrapper. 
    Requires Docker be installed, see https://www.docker.com/. 

    >>> _ = os.system('docker run --name muqbp -dit -p 4243:4243 linusseelinger/benchmark-muq-beam-propagation:latest > /dev/null')
    >>> import umbridge
    >>> dnb2 = DigitalNetB2(dimension=3,seed=7)
    >>> distribution = Uniform(dnb2,lower_bound=1,upper_bound=1.05)
    >>> model = umbridge.HTTPModel('http://localhost:4243','forward')
    >>> umbridge_config = {"d": dnb2.d}
    >>> um_bridge_integrand = UMBridgeWrapper(distribution,model,umbridge_config,parallel=False)
    >>> solution,data = CubQMCNetG(um_bridge_integrand,abs_tol=5e-2).integrate()
    >>> print(data)
    LDTransformData (AccumulateData Object)
        solution        [  0.      3.855  14.69  ... 898.921 935.383 971.884]
        comb_bound_low  [  0.      3.854  14.688 ... 898.9   935.362 971.862]
        comb_bound_high [  0.      3.855  14.691 ... 898.942 935.404 971.906]
        comb_flags      [ True  True  True ...  True  True  True]
        n_total         2^(11)
        n               [1024. 1024. 1024. ... 2048. 2048. 2048.]
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         0.050
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    UMBridgeWrapper (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     1
        upper_bound     1.050
    DigitalNetB2 (DiscreteDistribution Object)
        d               3
        dvec            [0 1 2]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()
    >>> _ = os.system('docker rm -f muqbp > /dev/null')
    >>> class TestModel(umbridge.Model):
    ...     def __init__(self):
    ...         super().__init__("forward")
    ...     def get_input_sizes(self, config):
    ...         return [1,2,3]
    ...     def get_output_sizes(self, config):
    ...         return [3,2,1]
    ...     def __call__(self, parameters, config):
    ...         out0 = [parameters[2][0],sum(parameters[2][:2]),sum(parameters[2])]
    ...         out1 = [parameters[1][0],sum(parameters[1])]
    ...         out2 = [parameters[0]]
    ...         return [out0,out1,out2]
    ...     def supports_evaluate(self):
    ...         return True
    >>> my_model = TestModel()
    >>> my_distribution = Uniform(
    ...     sampler = DigitalNetB2(dimension=sum(my_model.get_input_sizes(config={})),seed=7),
    ...     lower_bound = -1,
    ...     upper_bound = 1)
    >>> my_integrand = UMBridgeWrapper(my_distribution,my_model)
    >>> my_solution,my_data = CubQMCNetG(my_integrand,abs_tol=5e-2).integrate()
    >>> my_data
    LDTransformData (AccumulateData Object)
        solution        [-1.110e-16 -2.220e-16 -3.331e-16 -1.110e-16 -2.220e-16 -9.021e-17]
        comb_bound_low  [-7.639e-05 -8.012e-05 -7.965e-04 -5.369e-06 -3.118e-04 -9.537e-06]
        comb_bound_high [7.639e-05 8.012e-05 7.965e-04 5.369e-06 3.118e-04 9.537e-06]
        comb_flags      [ True  True  True  True  True  True]
        n_total         2^(10)
        n               [1024. 1024. 1024. 1024. 1024. 1024.]
        time_integrate  ...
    CubQMCNetG (StoppingCriterion Object)
        abs_tol         0.050
        rel_tol         0
        n_init          2^(10)
        n_max           2^(35)
    UMBridgeWrapper (Integrand Object)
    Uniform (TrueMeasure Object)
        lower_bound     -1
        upper_bound     1
    DigitalNetB2 (DiscreteDistribution Object)
        d               6
        dvec            [0 1 2 3 4 5]
        randomize       LMS_DS
        graycode        0
        entropy         7
        spawn_key       ()
    >>> my_integrand.to_umbridge_out_sizes(my_solution)
    [[-1.1102230246251565e-16, -2.220446049250313e-16, -3.3306690738754696e-16], [-1.1102230246251565e-16, -2.220446049250313e-16], [-9.020562075079397e-17]]
    >>> my_integrand.to_umbridge_out_sizes(my_data.comb_bound_low)
    [[-7.638736990567925e-05, -8.01234844033108e-05, -0.0007965080040872729], [-5.369077487201845e-06, -0.0003118494328260811], [-9.536743185422692e-06]]
    >>> my_integrand.to_umbridge_out_sizes(my_data.comb_bound_high)
    [[7.63873699054572e-05, 8.012348440286715e-05, 0.0007965080040866046], [5.3690774869798e-06, 0.000311849432825637], [9.536743185238812e-06]]

    References:

        [1] UM-Bridge documentation. https://um-bridge-benchmarks.readthedocs.io/en/docs/index.html
    """

    def __init__(self, true_measure, model, config={}, parallel=False):
        """
        See https://um-bridge-benchmarks.readthedocs.io/en/docs/umbridge/clients.html
        
        Args:
            true_measure (TrueMeasure): a TrueMeasure instance. 
            model (umbridge.HTTPModel): a UM-Bridge model 
            config (dict): config keyword argument to umbridge.HTTPModel(url,name).__call__
            parallel (int): If parallel is False, 0, or 1: function evaluation is done in serial fashion.
                Otherwise, parallel specifies the number of processes used by 
                multiprocessing.Pool or multiprocessing.pool.ThreadPool.
                Passing parallel=True sets processes = os.cpu_count().
        """
        import umbridge
        self.parameters = []
        self.true_measure = true_measure
        self.sampler = self.true_measure 
        self.model = model
        if not self.model.supports_evaluate(): raise ParameterError("UMBridgeWrapper requires model supports evaluation.")
        self.config = config
        self.parallel = parallel
        self.d_in_umbridge =  append(0,cumsum(self.model.get_input_sizes(self.config)))
        self.n_d_in_umbridge = len(self.d_in_umbridge)-1
        if self.d_in_umbridge[-1]!=self.true_measure.d:
            raise ParameterError("sampler dimension must equal the sum of UMBridgeWrapper input sizes.")
        self.d_out_umbridge = append(0,cumsum(self.model.get_output_sizes(self.config)))
        self.n_d_out_umbridge = len(self.d_out_umbridge)-1
        super(UMBridgeWrapper,self).__init__(
            dimension_indv = int(self.d_out_umbridge[-1]),
            dimension_comb = int(self.d_out_umbridge[-1]),
            parallel = self.parallel,
            threadpool = True)
    
    def g(self, t, **kwargs):
        n = len(t)
        y = zeros((n,self.d_indv[0]),dtype=float)
        for i in range(n):
            ti_ll = [t[i,self.d_in_umbridge[j]:self.d_in_umbridge[j+1]].tolist() for j in range(self.n_d_in_umbridge)]
            yi_ll = self.model.__call__(ti_ll,self.config)
            for j,yi_l in enumerate(yi_ll): y[i,self.d_out_umbridge[j]:self.d_out_umbridge[j+1]] = yi_l if len(yi_l)>1 else yi_l[0]
        return y
    
    def _spawn(self, level, sampler):
        return UMBridgeWrapper(
            true_measure = self.true_measure,
            model = self.model,
            config = self.config, 
            parallel = self.parallel)

    def to_umbridge_out_sizes(self, attr):
        """
        Convert a data attribute to UM-Bridge output sized list of lists. 
        
        Args:
            attr (ndarray): array of length sum(model.get_output_sizes(self.config))
        
        Return:
            list: list of lists with sub-list lengths specified by model.get_output_sizes(self.config)
        """
        return [attr[self.d_out_umbridge[j]:self.d_out_umbridge[j+1]].tolist() for j in range(self.n_d_out_umbridge)]



""" Plot functions used by Discrete Distribution and True Measure"""
import qmcpy as qp
from numpy import*
import os
def plot_proj(sampler, n = 64, d_horizontal = 1, d_vertical = 2,math_ind = True, marker_size = 5, figfac = 5, \
              fig_title = 'Projection of Samples', axis_pad = 0, want_grid = True, font_family = "sans-serif", \
                where_title = 1, **kwargs):
    """
    Args:
        sampler: the Discrete Distribution or the True Measure Object to be plotted
        n (int or list): the number of samples or a list of samples(used for extensibility) to be plotted. 
            Default value is 64
        d_horizontal (int or list): the dimension or list of dimensions to be plotted on the horizontal axes. 
            Default value is 1 (1st dimension).
        d_vertical (int or list): the dimension or list of dimensions to be plotted on the vertical axes. 
            Default value is 2 (2nd dimension).
        math_ind : setting it true will enable user to pass in math indices. 
            Default value is true, so user is required to pass in math indices.
        marker_size: the marker size in points**2(typographic points are 1/72 in.).
            Default value is 5.
        figfac: the figure size factor. Default value is 5.
        fig_title: the title of the figure. Default value is 'Projection of Samples'
        axis_pad: the padding of the axis so that points on the boundaries can be seen. Default value is 0.
        want_grid: setting it true will enable grid on the plot. Default value is true.
        font_family: the font family of the plot. Default value is "sans-serif".
        where_title: the position of the title on the plot. Default value is 1.
        **kwargs : Any extra features the user would like to see in the plots
    """
    try:
        import matplotlib.pyplot as plt
        dir_path = os.path.dirname(os.path.realpath(__file__))
        plt.style.use(os.path.join(dir_path, "../qmcpy.mplstyle"))
        from matplotlib import colors
    except:
        raise ImportError("Missing matplotlib.pyplot as plt, Matplotlib must be installed to run plot_proj function")
    plt.rcParams['font.family'] = font_family 
    n = atleast_1d(n)
    d_horizontal = atleast_1d(d_horizontal)
    d_vertical = atleast_1d(d_vertical)
    samples = sampler.gen_samples(n[n.size - 1])    
    d = samples.shape[1]
    fig, ax = plt.subplots(nrows=d_horizontal.size, ncols=d_vertical.size, figsize=(figfac*d_horizontal.size, figfac*d_vertical.size),squeeze=False)                    
    fig.tight_layout(pad=2)
    
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    for i in range(d_horizontal.size):         
        for j in range(d_vertical.size):
                n_min = 0
                for m in range(n.size):
                    n_max = n[m]  
                    if(d_horizontal[i] == d_vertical[j]):
                        ax[i,j].remove()  
                        break
                    if(math_ind is True):
                        x = d_horizontal[i] - 1
                        y = d_vertical[j] - 1
                        x_label_num = d_horizontal[i]
                        y_label_num = d_vertical[j]
                    else:
                        x = d_horizontal[i]
                        y = d_vertical[j]
                        x_label_num = d_horizontal[i] + 1
                        y_label_num = d_vertical[j] + 1
                    
                    if(isinstance(sampler,qp.DiscreteDistribution)):
                        ax[i,j].set_xlim([0-axis_pad,1+axis_pad])
                        ax[i,j].set_ylim([0-axis_pad,1+axis_pad])
                        ax[i,j].set_xticks([0,1/4,1/2,3/4,1])
                        ax[i,j].set_yticks([0,1/4,1/2,3/4,1])
                        ax[i,j].set_aspect(1)
                        if not want_grid:
                            ax[i,j].grid(False) 
                            ax[i,j].tick_params(axis='both', which='both', direction='in', length=5)
                        ax[i,j].grid(want_grid)
                        x_label = r'$x_{i%d}$'%(x_label_num)
                        y_label = r'$x_{i%d}$'%(y_label_num)
                    else:
                        x_label = r'$t_{i%d}$'%(x_label_num)
                        y_label = r'$t_{i%d}$'%(y_label_num)    
                    
                    ax[i,j].set_xlabel(x_label,fontsize = 15)
                    if(d > 1):
                        ax[i,j].set_ylabel(y_label,fontsize = 15)
                        y_axis = samples[n_min:n_max,y]
                    else:
                        y_axis = []
                        for h in range(n_max-n_min):
                            y_axis.append(0.5)
                    ax[i,j].scatter(samples[n_min:n_max,x],y_axis,s=marker_size,color=colors[m],label='n_min = %d, n_max = %d'%(n_min,n_max),**kwargs)
                    n_min = n[m]
    plt.suptitle(fig_title,fontsize = 20, y = where_title)
    #fig.text(0.55,0.55,fig_title, ha = 'center', va = 'center', fontsize = 20) %replaced by plt.suptitle
    return fig, ax


from ._stopping_criterion import StoppingCriterion
from ..accumulate_data import MLMCData
from ..discrete_distribution import IIDStdUniform
from ..discrete_distribution._discrete_distribution import IID
from ..true_measure import Gaussian
from ..integrand import MLCallOptions
from ..util import MaxSamplesWarning, ParameterError, MaxLevelsWarning, ParameterWarning
from numpy import *
from scipy.stats import norm
from time import time
import warnings


class CubMCML(StoppingCriterion):
    """
    Stopping criterion based on multi-level monte carlo.
    
    >>> mlco = MLCallOptions(IIDStdUniform(seed=7))
    >>> sc = CubMCML(mlco,abs_tol=.05)
    >>> solution,data = sc.integrate()
    >>> data
    MLMCData (AccumulateData Object)
        solution        10.450
        n_total         1213658
        levels          7
        n_level         [1.173e+06 2.369e+04 1.174e+04 3.314e+03 1.144e+03 4.380e+02 1.690e+02]
        mean_level      [1.006e+01 1.856e-01 1.053e-01 5.127e-02 2.699e-02 1.558e-02 7.068e-03]
        var_level       [1.958e+02 1.596e-01 4.603e-02 1.057e-02 2.978e-03 8.701e-04 2.552e-04]
        cost_per_sample [ 1.  2.  4.  8. 16. 32. 64.]
        alpha           0.936
        beta            1.870
        gamma           1.000
        time_integrate  ...
    CubMCML (StoppingCriterion Object)
        rmse_tol        0.019
        n_init          2^(8)
        levels_min      2^(1)
        levels_max      10
        theta           2^(-1)
    MLCallOptions (Integrand Object)
        option          european
        sigma           0.200
        k               100
        r               0.050
        t               1
        b               85
        level           0
    Gaussian (TrueMeasure Object)
        mean            0
        covariance      1
        decomp_type     PCA
    IIDStdUniform (DiscreteDistribution Object)
        d               1
        entropy         7
        spawn_key       ()

    Original Implementation:

        http://people.maths.ox.ac.uk/~gilesm/mlmc/#MATLAB

    References:
        
        [1] M.B. Giles. 'Multi-level Monte Carlo path simulation'. 
        Operations Research, 56(3):607-617, 2008.
        http://people.maths.ox.ac.uk/~gilesm/files/OPRE_2008.pdf.
    """

    def __init__(self, integrand, abs_tol=.05, alpha=.01, rmse_tol=None, n_init=256., n_max=1e10, 
        levels_min=2, levels_max=10, alpha0=-1., beta0=-1., gamma0=-1.):
        """
        Args:
            integrand (Integrand): integrand with multi-level g method
            abs_tol (float): absolute tolerance. Reset if supplied, ignored if not. 
            alpha (float): uncertainty level.
                If rmse_tol not supplied, then rmse_tol = abs_tol/norm.ppf(1-alpha/2)
            rel_tol (float): relative tolerance. Reset if supplied, ignored if not.
                Takes priority over absolute tolerance and alpha if supplied. 
            n_init (int): initial number of samples
            n_max (int): maximum number of samples
            levels_min (int): minimum level of refinement >= 2
            levels_max (int): maximum level of refinement >= Lmin
            alpha0 (float): weak error is O(2^{-alpha0*level})
            beta0 (float): variance is O(2^{-bet0a*level})
            gamma0 (float): sample cost is O(2^{gamma0*level})
        
        Note:
            if alpha, beta, gamma are not positive, then they will be estimated
        """
        self.parameters = ['rmse_tol','n_init','levels_min','levels_max','theta']
        if levels_min < 2:
            raise ParameterError('needs levels_min >= 2')
        if levels_max < levels_min:
            raise ParameterError('needs levels_max >= levels_min')
        if n_init <= 0:
            raise ParameterError('needs n_init > 0')
        # initialization
        if rmse_tol:
            self.rmse_tol = float(rmse_tol)
        else: # use absolute tolerance
            self.rmse_tol =  float(abs_tol) / norm.ppf(1-alpha/2)
        self.n_init = float(n_init)
        self.n_max = float(n_max)
        self.levels_min = float(levels_min)
        self.levels_max = float(levels_max)
        self.theta = 0.5
        self.alpha0 = alpha0
        self.beta0 = beta0
        self.gamma0 = gamma0
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        # Verify Compliant Construction
        allowed_levels = ['adaptive-multi']
        allowed_distribs = [IID]
        allow_vectorized_integrals = False
        super(CubMCML,self).__init__(allowed_levels, allowed_distribs, allow_vectorized_integrals)

    def integrate(self):
        """ See abstract method. """
         # Construct AccumulateData Object to House Integration Data
        self.data = MLMCData(self, self.integrand, self.true_measure, self.discrete_distrib,
            self.levels_min, self.n_init, self.alpha0, self.beta0, self.gamma0)
        t_start = time()
        while self.data.diff_n_level.sum() > 0:
            self.data.update_data()
            self.data.n_total += self.data.diff_n_level.sum()
            # set optimal number of additional samples
            n_samples = self._get_next_samples()
            self.data.diff_n_level = maximum(0, n_samples-self.data.n_level)
            # check if over sample budget
            if (self.data.n_total + self.data.diff_n_level.sum()) > self.n_max:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples, which would exceed n_max = %d.
                Stopping integration process.
                Note that error tolerances may no longer be satisfied""" \
                % (int(self.data.n_total), int(self.data.diff_n_level.sum()), int(self.n_max))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
            # if (almost) converged, estimate remaining error and decide 
            # whether a new level is required
            if (self.data.diff_n_level > 0.01*self.data.n_level).sum() == 0:
                range_ = arange(minimum(2.,self.data.levels-1)+1)
                idx = (self.data.levels-range_).astype(int)
                rem = (self.data.mean_level[idx] / 
                        2.**(range_*self.data.alpha)).max() / (2.**self.data.alpha - 1)
                # rem = ml(l+1) / (2^alpha - 1)
                if rem > sqrt(self.theta)*self.rmse_tol:
                    if self.data.levels == self.levels_max:
                        warnings.warn(
                            'Failed to achieve weak convergence. levels == levels_max.',
                            MaxLevelsWarning)
                    else:
                        self.data._add_level()
                        n_samples = self._get_next_samples()
                        self.data.diff_n_level = maximum(0., n_samples-self.data.n_level)
        # finally, evaluate multilevel estimator
        self.data.solution = (self.data.sum_level[0,:]/self.data.n_level).sum()
        self.data.levels += 1
        self.data.time_integrate = time() - t_start
        return self.data.solution,self.data
    
    def _get_next_samples(self):
        ns = ceil( sqrt(self.data.var_level/self.data.cost_per_sample) * 
                sqrt(self.data.var_level*self.data.cost_per_sample).sum() / 
                ((1-self.theta)*self.rmse_tol**2) )
        return ns
    
    def set_tolerance(self, abs_tol=None, alpha=.01, rmse_tol=None):
        """
        See abstract method. 
        
        Args:
            abs_tol (float): absolute tolerance. Reset if supplied, ignored if not. 
            alpha (float): uncertainty level.
                If rmse_tol not supplied, then rmse_tol = abs_tol/norm.ppf(1-alpha/2)
            rel_tol (float): relative tolerance. Reset if supplied, ignored if not.
                Takes priority over absolute tolerance and alpha if supplied.
        """
        if rmse_tol != None:
            self.rmse_tol = float(rmse_tol)
        elif abs_tol != None:
            self.rmse_tol = (float(abs_tol) / norm.ppf(1-alpha/2.))



from ._kernel_prod import _KernelProd
from ...discrete_distribution import IIDStdUniform
import numpy as np 
import itertools

class _KernelProdAutoGrad(_KernelProd):
    DEFAULTALPHA = np.nan
    def parsed_call(self, x1, x2, n1, n2, beta1s, beta2s, m1, m2, c1s, c2s, diag_only):
        try:
            import torch 
        except: 
            raise Exception("_AutoGradKernel requires torch for automatic differentiation")
        if diag_only:
            xmat1 = [x1[:,j] for j in range(self.d)]
            xmat2 = [x2[:,j] for j in range(self.d)]
            y = self.scale*torch.ones(n1,requires_grad=True,**self.ckwargs)
            grad_outputs = torch.ones(n1,dtype=torch.float,**self.ckwargs)
            v = torch.zeros(n1,requires_grad=False,**self.ckwargs)
        else:
            xmat1 = [self.npt.tile(x1[:,None,j],(1,n2)) for j in range(self.d)]
            xmat2 = [self.npt.tile(x2[None,:,j],(n1,1)) for j in range(self.d)]
            y = self.scale*torch.ones((n1,n2),requires_grad=True,**self.ckwargs)
            grad_outputs = torch.ones((n1,n2),dtype=torch.float,**self.ckwargs)
            v = torch.zeros((n1,n2),requires_grad=False,**self.ckwargs)
        for j in range(self.d):
            if self.npt==np:
                xmat1[j] = torch.from_numpy(xmat1[j])
                xmat2[j] = torch.from_numpy(xmat2[j])
            xmat1[j].requires_grad_()
            xmat2[j].requires_grad_()
        for j in range(self.d):
            y = y*self._1d_to_prod(xmat1[j],xmat2[j],self.lengthscales[j])
        for ell1,ell2 in itertools.product(range(m1),range(m2)):
            yc = y.clone()
            for j in range(self.d):
                for _ in range(beta1s[ell1,j]):
                    yc = torch.autograd.grad(yc,xmat1[j],create_graph=True,grad_outputs=grad_outputs)[0]
                for _ in range(beta2s[ell2,j]):
                    yc = torch.autograd.grad(yc,xmat2[j],create_graph=True,grad_outputs=grad_outputs)[0]
            if not self.requires_grad:
                yc = yc.detach()
            v = v+c1s[ell1]*c2s[ell2]*yc#.detach()
        if self.npt==np: 
            v = v.detach().cpu().numpy().astype(np.float64) 
        return v

class KernelGaussian(_KernelProdAutoGrad):
    """ Gaussian kernel 
    
    >>> import torch
    >>> d = 3
    >>> x = torch.from_numpy(IIDStdUniform(d,seed=7).gen_samples(8))
    >>> k = KernelGaussian(d,torchify=True)
    >>> with np.printoptions(precision=2):
    ...     k(x,x)
    tensor([[1.0000, 0.7980, 0.7412, 0.9196, 0.8771, 0.7125, 0.6546, 0.7469],
            [0.7980, 1.0000, 0.7575, 0.7362, 0.6654, 0.6713, 0.7642, 0.8466],
            [0.7412, 0.7575, 1.0000, 0.7915, 0.7102, 0.9872, 0.7444, 0.9120],
            [0.9196, 0.7362, 0.7915, 1.0000, 0.7092, 0.7737, 0.5166, 0.8509],
            [0.8771, 0.6654, 0.7102, 0.7092, 1.0000, 0.7067, 0.7925, 0.5816],
            [0.7125, 0.6713, 0.9872, 0.7737, 0.7067, 1.0000, 0.6982, 0.8583],
            [0.6546, 0.7642, 0.7444, 0.5166, 0.7925, 0.6982, 1.0000, 0.6141],
            [0.7469, 0.8466, 0.9120, 0.8509, 0.5816, 0.8583, 0.6141, 1.0000]],
           dtype=torch.float64)
    """
    def _1d_to_prod(self, xmat1j, xmat2j, lengthscale):
        import torch 
        return torch.exp(-(xmat1j-xmat2j)**2/(2*lengthscale))


from .gaussian import Gaussian
from ..discrete_distribution._discrete_distribution import DiscreteDistribution
from ._true_measure import TrueMeasure
from ..discrete_distribution.lattice.lattice import Lattice
from ..util import DimensionError, ParameterError
from numpy import *
from scipy.special import kv, gamma

class Matern(Gaussian):
    """
    A normal measure using a Matern kernel as the covariance matrix.
    >>> mean = full(31, 1.1)
    >>> x_values = arange(31.0) / 30.0 #[0, 1/30, ..., 1]
    >>> m = Matern(Lattice(dimension=31, seed=7), x_values, length_scale = 0.5, variance = 0.01, mean=mean)
    >>> m
    Matern (TrueMeasure Object)
        mean            [1.1 1.1 1.1 ... 1.1 1.1 1.1]
        covariance      [[0.01  0.01  0.01  ... 0.002 0.002 0.001]
                        [0.01  0.01  0.01  ... 0.002 0.002 0.002]
                        [0.01  0.01  0.01  ... 0.002 0.002 0.002]
                        ...
                        [0.002 0.002 0.002 ... 0.01  0.01  0.01 ]
                        [0.002 0.002 0.002 ... 0.01  0.01  0.01 ]
                        [0.001 0.002 0.002 ... 0.01  0.01  0.01 ]]
        decomp_type     PCA
    >>> m.gen_samples(1)
    array([[1.22818861, 1.22050757, 1.21522052, 1.2144905 , 1.21449756,
            1.22499156, 1.24161076, 1.25371086, 1.26118833, 1.2718846 ,
            1.27662595, 1.28237625, 1.28688038, 1.28466947, 1.28328019,
            1.27822864, 1.26792407, 1.25583031, 1.25052001, 1.24470164,
            1.2384988 , 1.2308771 , 1.22744825, 1.2275582 , 1.218964  ,
            1.2136328 , 1.2141693 , 1.21792116, 1.21125532, 1.19532303,
            1.17649556]])
    >>> points = array([[5, 4], [1, 2], [0, 0]])
    >>> mean = full(3, 1.1)
    >>> m2 = Matern(Lattice(dimension = 3,seed=7), points, length_scale = 4, nu = 2.5, variance = 0.01, mean=mean)
    >>> m2
    Matern (TrueMeasure Object)
        mean            [1.1 1.1 1.1]
        covariance      [[0.01  0.005 0.002]
                        [0.005 0.01  0.008]
                        [0.002 0.008 0.01 ]]
        decomp_type     PCA
    >>> from sklearn import gaussian_process as gp  #checking against scikit's Matern
    >>> kernel2 = gp.kernels.Matern(length_scale = 4, nu=2.5)
    >>> cov2 = 0.01 * kernel2.__call__(points)
    >>> isclose(cov2, m2.covariance)
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]])
    >>> x_values = array([0, 1, 2])
    >>> m3 = Matern(Lattice(dimension = 3,seed=7), x_values, length_scale = 0.5, nu = 3.5, variance = 0.01, mean=mean, decomp_type = 'Cholesky')
    >>> m3
    Matern (TrueMeasure Object)
        mean            [1.1 1.1 1.1]
        covariance      [[1.000e-02 1.378e-03 3.432e-05]
                        [1.378e-03 1.000e-02 1.378e-03]
                        [3.432e-05 1.378e-03 1.000e-02]]
        decomp_type     CHOLESKY

     References:

        [1] scikit-learn developers, Matern kernel. 
        https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html.
        Accessed 2023.

        [2] Abramowitz and Stegun (1972). Handbook of Mathematical Functions with Formulas, Graphs, 
        and Mathematical Tables. ISBN 0-486-61272-4. (Accessed through Wikipedia, 
        https://en.wikipedia.org/wiki/Mat%C3%A9rn_covariance_function.)
    """

    def __init__(self, sampler, points, length_scale = 1.0, nu = 1.5, variance = 1.0, mean = [], decomp_type='PCA'):
        """
        Matern kernel: calculates covariance over a metric space based only on the distance between points.
        More information can be found at [1].
        
        Args:
            sampler (DiscreteDistribution/TrueMeasure): A 
                discrete distribution from which to transform samples or a
                true measure by which to compose a transform. 
            points (ndarray): The positions of points on a metric space. The array
                should be of shape n x d: n points, each of dimension d.
            length_scale (float): Determines "peakiness", or how correlated 
                two points are based on their distance.
            nu (float): The "smoothness" of the Matern function. e.g. nu=1.5
                implies a once-differentiable function, while nu=2.5 implies twice
                differentiability. Meanwhile, when nu=0.5, the Matern kernel equals
                the squared exponential kernel, while as nu approaches inf, it equals 
                the RBF kernel. Note that nu values not in [0.5, 1.5, 2.5, inf] 
                will be ~10x slower to run.
            variance (float): The variance (or the diagonal elements of the
                covariance matrix) of the distribution at each point. Implemented
                as a scaling factor of the Matern covariance. The default 
                value is 1.0.
            mean (float): mu for Normal(mu,sigma^2)
            decomp_type (str): method of decomposition either  
                "PCA" for principal component analysis or 
                "Cholesky" for cholesky decomposition.
        """
        if not (isinstance(sampler, DiscreteDistribution) or isinstance(sampler, TrueMeasure)):
            raise ParameterError("sampler input should either be a DiscreteDistribution or TrueMeasure.")
        if not (isinstance(points, ndarray)):
            raise ParameterError("Must pass in a points ndarray.")
        if not (sampler.d == len(mean) and points.shape[0] == len(mean)):
            raise DimensionError("The lengths of the sampler and mean array and the number of points must all be equal.")
        
        if len(points.shape) == 1: #one dimensional points array, 1 x N
            points = atleast_2d(points).T
        self.points = points
        self.length_scale = length_scale
        self.nu = nu
        self.variance = variance
        
        # See [1], [2] for Matern formula
        # Replicating scikit-learn's Matern kernel using hard-coded cases, scipy.special.kv
        dimension = len(mean)
        rho = length_scale
        covariance = zeros((dimension, dimension))
        
        for i in range(dimension):
            for j in range(dimension):
                d = linalg.norm(points[i] - points[j])
                if d == 0:
                    covariance[i][j] = 1 #convergent value
                elif isclose(nu, 0.5):
                    covariance[i][j] = exp(-1 * d / rho)
                elif isclose(nu, 1.5):
                    covariance[i][j] = (1 + sqrt(3) * d / rho) * exp(-1 * sqrt(3) * d / rho)
                elif isclose(nu, 2.5):
                    covariance[i][j] = ((1 + sqrt(5) * d / rho + 5 * d ** 2 / (3 * rho ** 2)) * 
                                        exp(-1 * sqrt(5) * d / rho))
                else:
                    k = sqrt(2 * nu) * d / rho
                    covariance[i][j] = 2 ** (1 - nu) / gamma(nu) * k ** nu * kv(nu, k)
                """ this doesn't produce a positive definite matrix, so it isn't accepted in Gaussian
                elif nu == inf:
                    covariance[i][j] = exp(-1 * d ** 2 / (2 * rho ** 2))
                """
                covariance[i][j] = covariance[i][j] * variance
        
        #print(covariance)
        super().__init__(sampler, mean=mean, covariance=covariance, decomp_type=decomp_type)

    def _spawn(self, sampler):
        return Matern(sampler, self.points, length_scale=self.length_scale, nu=self.nu, variance=self.variance, 
                      mean=self.mean, decomp_type=self.decomp_type)
        
